/**************
This file deals with a global array of images for the laser.
This array will extend itself (5 by 5) each time it is needed.
**************/


#include <stdio.h>
#include <math.h>
#ifdef VXWORKS
#define M_PI 3.141592653589793
#endif
#include <stdlib.h>


#include "lasdef.h"
#include "las_prot.h"
#include "dbliste.h"
#include "image_mgr_p.h"
#include "las_io_p.h"
#include "mat.h"

static igla **Glob_images = NULL;
static long Nb_Images = 0;

/* ---------------------------------------------------------------------- */
/* PROTOTYPES OF LOCAL FUNCTIONS */

/*******
creates or extend the array of images
*******/
static long init_image_array(long n);


/**********
  Creates and fill the image n with the current perception
**********/

static long LasCreate_image(long nbp,long *laser_val);



#ifdef UNUSED
static long CamCreate_image(long nbp, long *laser_val );
#endif /* UNUSED */ 
 
/**************
Adds the points of ig1 to ig2.
It assumes the center of  
the robot didn't move between perceptions.
We get rid of the overlapping by not taking into 
account the points of ig1 overlaping (in angle)
the points of ig2.

Warning: suppose the  perceptions made in trigonometric 
order.
**************/


static long Append_static_image(igla *ig1,igla *ig2);


/*********
adds several local images taken in the angular order from the same 
point; allows the 360 degree scanning
*********/ 

static long Concat_images(PDBL *List_im);
 
/* END PRIVATE PROTOTYPES */
/* ---------------------------------------------------------------------- */

#define deg2rad(x) (3.141592653589793 * (x) /180.)

#define rad2deg(x) ((x)*180/3.141592653589793)

static long started = 0;


/*******
creates or extend the array of images
*******/

static long init_image_array(long n)
    

{
  long i;
  igla **ig;

  
  ig = Glob_images;

  if((Glob_images = (igla **)malloc((n + Nb_Images)* sizeof(igla *))) == NULL){
      printf("pb for allocating the array of images\n"); 
      printf("Glob_images conserved \n");
      Glob_images = ig;
      return(0);
    }
  
  for(i = 0; i < Nb_Images; i++){
    Glob_images[i] = ig[i];
  }
  for(i = Nb_Images; i < Nb_Images + n; i++)
    Glob_images[i] = NULL;

  Nb_Images += n;
  /* printf(" %d images created \n",Nb_Images); */
  if(ig)
    free(ig);
  return(1);
}
  
void Gfree_image(long n)
     

{
  if(n >= 0){
    if(Glob_images[n] != NULL){
      Imgl_free(Glob_images[n]);
      Glob_images[n] = NULL;
    }
  }

}
  
long Get_Image(void)
{

  long i;

  for(i = 0; i < Nb_Images; i++){
    if(Glob_images[i] == NULL){
      /* printf(" allocating image %d\n",i); */
      Glob_images[i] = Imgl_cree();
      return(i);
    }
  }
  init_image_array(5);
  /* printf(" allocating image %d\n",Nb_Images - 5); */
  Glob_images[Nb_Images - 5] = Imgl_cree();
  return(Nb_Images - 5);
}

/**********
  Creates and fill the image n with the current perception
**********/

static long LasCreate_image(long nbp,long *laser_val)

{
  long n;

  n = Get_Image();
  Imgl_cart_fill(Glob_images[n],nbp,laser_val);
  return(n);

}

#ifdef UNUSED
static long CamCreate_image(long nbp, long *laser_val )
     
{
  long n;

  n = Get_Image();
  Imgl_cam_fill(Glob_images[n],nbp,laser_val);
  return(n);

}
#endif /* UNUSED */

long SegmCreate_image(long nbp, long *laser_val)
     
{
  long n1;
 
  n1 = LasCreate_image(nbp,laser_val);

  segmentation(PtoIm(n1),1); 

  return(n1);

}

/************
  write the data segment image n
  in a file. call to wrt_file_sg_c
  (las_io.c);
*************/

  
long write_file_sg(long n)
{
 
  static char file[100];
  printf("file name :\n");

  scanf("%s",file);
  
  if(n >= 0)
      return(wrt_file_sg_c(PtoIm(n),file));
  return(1);
}

/************
  read a data file
  in a segment image;
  returns an index on the segment image
*************/

long read_file_sg(const char *file)
{
  long im;
  
  
  im = Get_Image();
  
  if(rd_file_sg_c(PtoIm(im),file) == -1)
    return(-1);
  
  printf("image number %ld: %d points and %ld segments \n",
	 im,PtoIm(im)->nbp,PtoIm(im)->nsg);
  return(im);
  
}

    
/*******************
  Returns the actual pointer on the image
  in position n in the global array
*******************/

igla *PtoIm(long n)
     
{
  return(Glob_images[n]);
}

/************************
Interface to Set_pos_image in las_struct.c 
Sets the global position of the image indexed by "image"
to be "pos"
*************************/

void SetAbsPosImage(int image, double *pos)
{

  Set_pos_image(PtoIm(image),pos[0],pos[1],pos[2],pos[3]);

}
/********************************
Interface to set_encoders() in odom_sim.c
Sets the current value of encoders to be the robot encoder values
at the time of the perception of "im"
*********************************/

/*
void set_encoders_from_image(im)
{

set_encoders(PtoIm(im)->enc, PtoIm(im)->var);

}
*/


/****************
  copies an image of points in the array
  laser_point  (must be allocated).

  laser_point[0] = number 0f segments;

  x,y;

  unit: millimeters
********************/

long Copy_image_point(long im,long *laser_point, int maxarraysize)
{
  p2d *pt;
  igla *ig;
  
  long i = 0;
  if(im < 0){
    printf("Error: can't access to negative image number %ld\n",im);
    return(-1);
  }
  if((ig = PtoIm(im)) == NULL){
    printf("Error in copying image number %ld: no structure allocated !\n",im);
    return(-1);
  }  

  if (ig->nbp > maxarraysize) {
    printf("Copy_image_point Error in copying image number %ld: too many points (%d to be copied in a %d tab size) !\n",im, ig->nbp, maxarraysize);
    return(-1);
  }  
    
  laser_point[0] = ig->nbp;
  pt = ig->tbp;
  for(i = 0; i <ig->nbp; i++){
    
   
    laser_point[2*i + 1] = (long)((pt+i)->x  * 1000.);
    laser_point[2*i + 2] =(long)( (pt+i)->y * 1000.);
        
  }

  return(1);

}




/****************
  copies an image segment in the array
  laser_segment  (must be allocated).

  laser_segment[0] = number 0f segments;

  x1,y1,x2y2;

  unit: millimeters

  orientation : free-space on the right side

********************/

long Copy_image_segment(long im,long *laser_segment, int maxarraysize)
{
  igla *ig;
  PDBL *l;
  Sgl *s;
  long i = 0;
  if(im < 0){
    printf("Copy_image_segment Error: can't access to negative image number %ld\n",im);
    return(-1);
  }
  if((ig = PtoIm(im)) == NULL){
    printf("Copy_image_segment Error in copying image number %ld: no structure allocated !\n",im);
    return(-1);
  }  

  if (ig->nsg > maxarraysize) {
    printf("Copy_image_segment Error in copying image number %ld: too many segments (%ld to be copied in a %d tab size) !\n",im, ig->nsg, maxarraysize);
    return(-1);
  }  
    
  laser_segment[0] = ig->nsg;

  for(l = ig->lsgl; l != NULL; l = dls(l)){
    s = (Sgl *)dlel(l);
   
    laser_segment[4*i + 1] = (long)(s->first.x  * 1000.);
    laser_segment[4*i + 2] =(long)( s->first.y * 1000.);
    laser_segment[4*i + 3] = (long)(s->last.x * 1000.);
    laser_segment[4*i + 4] = (long)(s->last.y * 1000.);
    i++;
  }

  return(1);

}




/***
  same as above but filter out segment with
  ->flg != 0. 
****/

long Copy_image_segment2(long im,long *laser_segment, int maxarraysize)
{
  igla *ig;
  PDBL *l;
  Sgl *s;
  long i = 0;

  if(im < 0){
    printf("Copy_image_segment2 Error: can't access to negative image number %ld\n",im);
    return(-1);
  }
  if((ig = PtoIm(im)) == NULL){
    printf("Copy_image_segment2 Error in copying image number %ld: no structure allocated !\n",im);
    return(-1);
  }  

  for(l = ig->lsgl; l != NULL; l = dls(l)){
    s = (Sgl *)dlel(l);
    if (s->flg == 0) { /* good segments */
      laser_segment[4*i + 1] = (long)(s->last.x*1000.);
      laser_segment[4*i + 2] = (long)(s->last.y*1000.);
      laser_segment[4*i + 3] = (long)(s->first.x*1000.);
      laser_segment[4*i + 4] = (long)(s->first.y*1000.);
      i++;
      if (i == maxarraysize) {
	laser_segment[0] = i;
	printf("Copy_image_segment2 Error in copying image number %ld: too many segments (%ld to be copied in a %d tab size) !\n",im, ig->nsg, maxarraysize);
	return(-1);
      }  
    }
  }
  laser_segment[0] = i;

  return(1);

}

/********************************
  look at the image im to know if it is segmented
  if not, calls the segmentation
  ********************************/

int SegmentIfNot(int im)
{
  igla *ig;

  if ( im < 0)
    return (im);

  if((ig = PtoIm(im)) == NULL)
    return(im);

  if(ig->tbp == NULL || ig->nbp == 0 || ig->lsgl != NULL)
    return(im);
  
  segmentation(ig,1);

}

  
/*************
 Creates a global image of laser points
 from an array of laser_val_array[3] snap shots.
 taken from the position
 x = laser_val_array[0], y = laser_val_array[1], th = laser_val_array[2].
 laser_val_array[4] = turret angle of the first snap shot
 laser_val_array[5] = number of point of the first snap shot
 then come the first snap shot pixel values
 etc ....
 defined in LaserAc.c
*************/

long LasScanSegmentation(long *laser_val_array)
     
{
  
  long im;

  
  if(!started){

    /**********
      Initializing the parameters for future
      segmentation. Reading the calibration file
      Laser.cal
    ***********/
    
     Read_init_calib();
     conf_sg_host(60,6,30,150);
    
    started = 1;
  }

  
  /********
    transform the array laser_val_array of
    n snap shots in pixels
    into one single image
    in cartesian coordinates
    im = index on the resulting image
  ********/
    
  im = LasScanningTransformation(laser_val_array);

  /******
    segmentation of the image
  ******/
  segmentation(PtoIm(im),1);
 
    if(im >= 0 ){
      if(PtoIm(im) != NULL){
	printf("Image number %ld : %d points + %ld segments \n",
	       im,PtoIm(im)->nbp,PtoIm(im)->nsg); 
	return(im);
    }
  }
  
  
  return(-1);
}    

/*************
 Creates a global image of laser points
 from an array of laser_val_array[3] snap shots.
 taken from the position
 x = laser_val_array[0], y = laser_val_array[1], th = laser_val_array[2].
 laser_val_array[4] = turret angle of the first snap shot
 laser_val_array[5] = number of point of the first snap shot
 then come the first snap shot pixel values
 etc ....
 
*************/

long LasScanningTransformation(long *laser_val_array)
  
{
  long NbSnapShots;
  long Concat;
  PDBL *Lper = NULL,*tmp_l;
  long *tbper,i;
  long nbp = 0, nbptot = 0;
  long begin;
  long Encoders[4];

  NbSnapShots = laser_val_array[3];

  tbper = (long *)malloc(NbSnapShots * sizeof(long)); 

  /* filling an array for encoders values */

  Encoders[0] = laser_val_array[0];
  Encoders[1] = laser_val_array[1];
  Encoders[2] = laser_val_array[2];

  for(i = 0; i < NbSnapShots; i++){
    begin = 2*i + nbptot + 5;
    nbp = laser_val_array[begin];
    nbptot += nbp;


    /*******
      transforming the pixel image into cartesian coordinates
    *******/

    tbper[i] = LasCreate_image(nbp,&laser_val_array[begin+1]);

    /*****
      setting encoder values for this snap shot
    *****/

    Encoders[3] = laser_val_array[begin-1];

    Set_enc_image(PtoIm(tbper[i]),Encoders);

    /***
      storing in a list
    ***/

    if(i == 0)/* memorizing the head of the list */
      tmp_l =  Lper = dlbcons(NULL,tbper);
    else /* adding at the end for keeping the order */
      tmp_l = dlbcons(tmp_l,tbper + i);
  }
  /****
    joining the snap shots in one image 
  ****/
  Concat = Concat_images(Lper);

  /** 
    freeing the intermediate list
  **/

  for(tmp_l = Lper; tmp_l != NULL; tmp_l = dls(tmp_l)){
    
    Gfree_image(*((long *)dlel(tmp_l)));

  } 
  dlib_list(Lper);
  free(tbper);

  /***
    returning the index on the final image
  ***/
  return(Concat);

}



/*********
adds several local images taken in the angular order from the same 
point; allows the 360 degree scanning
*********/ 

static long Concat_images(PDBL *List_im)
 

{

  igla *ig,*igres;
  PDBL *l;
  long ntot = 0,numig=0;

  

  
  /* total number of points */

  for(l = List_im; l != NULL; l = dls(l)){
    ig = Glob_images[*((long *)dlel(l))];
    numig++;
    ntot += ig->nbp;
  }

  numig = Get_Image();

  igres = Glob_images[numig];

  igres->nbp = 0;

  
  ig = Glob_images[*((long *)dlel(List_im))];
  

  igres->x = ig->x;
  igres->y = ig->y;
  igres->cs = ig->cs;
  igres->sn = ig->sn;
  igres->cst = ig->cst;
  igres->snt = ig->snt;
  igres->coosys = 'S';
  mx_copy(ig->enc,igres->enc,4,1);
  igres->tbp = (p2d *)malloc(ntot*sizeof(p2d));

  

  for(l = List_im; l != NULL; l = dls(l)){
    ig = Glob_images[*((long *)dlel(l))];
    if(Append_static_image(ig,igres) == -1)
      break;
  }

  
  return(numig);
}




 
/**************
Adds the points of ig1 to ig2.
It assumes the center of  
the robot didn't move between perceptions.
We get rid of the overlapping by not taking into 
account the points of ig1 overlaping (in angle)
the points of ig2.

Warning: suppose the  perceptions made in trigonometric 
order.
**************/


static long Append_static_image(igla *ig1,igla *ig2)


{



static double r1[4],vaux1[3],vaux2[3];
double ag,last_ag;
static double ag_tot = 0;
long n1,n2,i,j,nn,off;
p2d *p1,*p2;





ag = norm_ang(ig1->enc[3] - ig2->enc[3],-M_PI);

r1[0] = cos(ag);  
r1[1] = -sin(ag);
r1[2] = -r1[1];
r1[3] = r1[0];

/*******
  adding
*******/

n1 = ig1->nbp;
n2 = ig2->nbp;
p1 = ig1->tbp;
p2 = ig2->tbp;

/*******
  getting the last  angle of ig2:
*******/
if(ig2->nbp >= 1)
  last_ag = ig2->tbp[ig2->nbp-1].o;
else{
  last_ag = 1000;
  ag_tot = 0;
}
/*******
  Looking for the first good angle of ig1
*******/

i = 0;
if(last_ag < 1000)
  while(norm_ang(p1[i].o - last_ag + ag,-M_PI) < 0 && i < ig1->nbp)
  i++;


  
off = i;
nn = ig1->nbp - off;
j = off;

for(i = n2; i < nn + n2; i++){

  
  (p2 + i)->o = (p1 + j)->o + ag;

  /***
    checking we don't go more than 360 degrees
  ***/
  if(i > 0)
    {
      ag_tot += norm_ang((p2+i)->o - (p2 + i -1)->o,-M_PI);
      if(ag_tot > 2*M_PI){
	ig2->nbp = i;
	return(-1);
      }
    }
  (p2 + i)->x = (p1 + j)->x * r1[0] + (p1 + j)->y * r1[1] ;
  (p2 + i)->y = (p1 + j)->x * r1[2] + (p1 + j)->y * r1[3] ;
  (p2 + i)->r = (p1 + j)->r;

  (p2 + i)->vr = (p1 + j)->vr;
  (p2 + i)->vo = (p1 + j)->vo;
  (p2 + i)->vro = (p1 + j)->vro;


  vaux1[0] = (p1+j)->vx; vaux1[1] = (p1+j)->vxy; vaux1[2] = (p1+j)->vy;

  mxfs_mmulttr(r1,vaux1,vaux2,2,2,2,2);

  (p2 + i)->vx = vaux2[0]; (p2+i)->vxy = vaux2[1]; (p2+i)->vy = vaux2[2];
  
  j++;
}

ig2->nbp = i;
return(1);
}















