/******************
  creating, deleting, initializing structures
  of point
******************/




#include <stdio.h>
#include <math.h>
#ifdef VXWORKS
#define M_PI 3.141592653589793
#endif
#include <stdlib.h>
#include "dbliste.h"
#include "las_prot.h"
#include "mat.h"



static FILE *ptr;


/* laser parameters (see lasdef.h) */

static long Nbpt; 
/* number of point of the last acquisition */


static double MaxPtDist = 120.00;
 
/* maximum distance of a point */
/* useful if points are known to be
   very innacurate when far away */


/*********
things for variance:
**********/


/* variance corresponding to a uniform
   error of 1 on each pixel ; */

static double vpix_x = 0.08333333333; 

static double vpix_y = 0.08333333333;

/* intrinsic parameters of the camera/physical
   space. Defined relatively to a virtual point Xv */

static double CamParam[8]; 

/* coordinates of Xv relatively to the center of
				the robot */

static double TransRobot[2]; 

/* variance matrix on CamParam[] */

static double VarParam[36]; 




/****************
  local variables
*****************/


/********
variance stuff (matrix plus jacobians) 
*******/

static double Jxi[4],Jpar[16];

static double Vxpi[3], Vxpp[3];



/***************
 Calibrating file
*****************/

static char File[50] = "laser.cal";



/***********************************************************
Structures allocation and destruction function
***********************************************************/




/***************
 * FUNCTION:     Pt_cree
 * PURPOSE:      creates an array of n points
                 and returns its pointer
 * ARGUMENTS:    long n :
 * ALGORITHM:    
 * RETURN:       p2d *
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

p2d *Pt_cree(long n)

{

  p2d *p;

  if( n == 0)
    return (NULL);

  if( (p = (p2d *)malloc(n * sizeof(p2d))) == NULL)
	  printf(" point allocation failed !\n");

  return(p);
}

/***************
 * FUNCTION:     Sgl_cree
 * PURPOSE:      creates a segment structure and returns 
                 its pointer. 
		 sets the extremity connections to NULL
 * ARGUMENTS:    None
 * ALGORITHM:    
 * RETURN:       Sgl *
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

Sgl *Sgl_cree(void)
{
  Sgl *s;
  if( (s = (Sgl *)malloc(sizeof(Sgl))) == NULL)
	  printf(" segment allocation failed !\n");
  s->slast = NULL; s->sfirst = NULL;
  s->flg = 0;
  return(s);
}


/***************
 * FUNCTION:     Sgl_free
 * PURPOSE:      free segment structure
 * ARGUMENTS:    Sgl *s :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void Sgl_free(Sgl *s)

{
  free(s);
}


/***************
 * FUNCTION:     Imgl_cree
 * PURPOSE:      creates and initialize a structure image laser
                 array of points is set to NULL
		 list of segments is setto NULL
		 coordinate system is set to (0,0,0,0) and 
		 defined as absolute by default (coosys = 'A')
		 iniializes the varaiance of the robot position
		 to 0.
 * ARGUMENTS:    None
 * ALGORITHM:    
 * RETURN:       igla *
 * SIDE EFFECT:  
 * CALLS:        mx_init/matrix.c
 * CALLED BY:    
 ***************/

igla *Imgl_cree(void)
{
  igla *ig;

  if( (ig = (igla *)malloc(sizeof(igla))) == NULL){
    printf(" image laser allocation failed !\n");
    return(NULL);
  }

  ig->nbp = 0;
  ig->tbp = NULL;
  ig->nsg = 0;

  ig->lsgl = NULL;
  ig->x = 0.;
  ig->y = 0.;
  ig->cs = 1.;
  ig->sn = 0.;
  ig->cst = 1.;
  ig->snt = 0.;
  ig->coosys = 'A';
  mxs_init(ig->var,3,3,0.);
  return(ig);
}
    

/***************
 * FUNCTION:     Imgl_free
 * PURPOSE:      free the image *ig and all of its dynamic components
                 array of points and list of segments
 * ARGUMENTS:    igla *ig :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        dlib_list_el()/dbliste.c
 * CALLED BY:    
 ***************/

void Imgl_free(igla *ig)
  

{
  if (ig == NULL)
    return;
  free(ig->tbp);
  dlib_list_el(ig->lsgl, Sgl_free);
  free(ig);
}


/**************************************************************
Data Acquisition Functions
   Works with a set of two functions
   -image structure creation and filling;
   -point array creation and filling

Pixel acquisition:

    Imgl_cam_fill
    ReadCamPer

Cartesian coordinates acquisition
    Imgl_cart_fill
    ReadCamCart

**************************************************************/



/***************
 * FUNCTION:     SetMaxPtDist
 * PURPOSE:      set the paramter MaxPtDist (tenth of inch) 
                 all points further than MaxPtDist from the robot center
		 will be discarded during laser data acquisition process
		 in ReadCamCart().
 * ARGUMENTS:    long maxdt :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void SetMaxPtDist(long maxdt)
{
  MaxPtDist = ((double)maxdt)/10.;
} 


/******
Create an array of points and 
fill it with the perception directly from the camera (pixels)
You should have done Set_ls_pix(482) before calling this function
*******/


/***************
 * FUNCTION:     ReadCamPer
 * PURPOSE:      creates an array of points "p2d" and fill it with 
                 the values in the laser value array "laser_val".
		 It is intended only for direct pixel reading.
		 Variance matrices are computed assuming a uniform distribution
		 one pixel wide in "y" coordinate and a gaussian distribution
		 of variance 1 in x coordinate. 
		 x represents the laser value itself
		 y represents the scan line of the camera.
		 Typical use after a gs() call:
		      ReadCamPer(Laser[0],&Laser[1]);
		 The function also computes polar transformations 
		 WARNING: this function assumes you took 482 points and is mainly
		 intended for calibration purposes. 
		 
 * ARGUMENTS:    long nbp :
 *               long *laser_val :
 * ALGORITHM:    
 * RETURN:       p2d *
 * SIDE EFFECT:  
 * CALLS:        Pt_cree() , point_inc_xyro()/las_unc.c
 * CALLED BY:    Imagl_cam_fill()
 ***************/


p2d *ReadCamPer(long nbp,long *laser_val)
{
  int i,k=0;

  double x,y,d,ag;
  p2d *pt;
  
  
  

  Nbpt = nbp;
  
  pt = Pt_cree(Nbpt);
  for(i = 0; i < Nbpt; i++){
    
    if(laser_val[i]  > 0 && laser_val[i] < 60000){
      y = (pt + k)->y =  i - 241.5;
      x = (pt + k)->x = laser_val[i] - 255.5;
      
      (pt + k)->r = d = sqrt(x*x + y*y);
      (pt + k)->o = ag = atan2(y,x);
      
      (pt + k)->vx = 12*vpix_x;
      (pt + k)->vy = vpix_y;
      (pt + k)->vxy = 0.;
      point_inc_xyro(pt+k);
      k++;
    }
      
    
  }
  Nbpt = k;
  return(pt);
}
  

/***************
 * FUNCTION:     Imgl_cam_fill
 * PURPOSE:      fills the igla structure "*ig" with the
                 the laser data contained in "*laser_val" directly.

		 
		 The coordinate system is set to be absolute: 
		 ig->coosys = 'A'
		 In case you use this with the robot directly, the 
		 functions "Set_pos_image" and "Set_enc_image" allow
		 you to save encoder values in the *igla structure.
		 Here, all values are set to 0 (encoder position
		 + encoder angles + absolute position + absolute angles)

		 ig->x, ig->y = x and y absolute  of the robot
		               at perception time
		 ig->cs,ig->sn = cos and sine of encoder sterring angle
		 ig->cst, ig->snt =  cos and sine of encoder turret angle
		 ig->enc = rob_enc : x,y,st,tu
 
		 
		 All fields ->x,->y,->cs etc are intended to contain absolute
		 coordinates when different from the robots' starting point.
		 
		 ig->enc[4] should always keep memory
		 of the encoder values at the time the perception was made.
		 
 * ARGUMENTS:    igla *ig :
 *               long nbp :
 *               long *laser_val :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        ReadCamPer(), mx_init/matrix.c
 * CALLED BY:    
 ***************/

void Imgl_cam_fill(igla *ig,long nbp,long *laser_val)
    
{
  
  ig->tbp = ReadCamPer(nbp,laser_val);
  ig->nbp = Nbpt;
  ig->x = 0.;
  ig->y = 0.;
  ig->cs = 1.;
  ig->sn = 0.;
  ig->cst = 1.;
  ig->snt = 0.;
  mx_init(ig->enc,4,1,0.);
  ig->coosys = 'A';

}

/***************
 * FUNCTION:     ReadCamCart
 * PURPOSE:      Creates an array of points
                 Fill it with the current perception in physical space
		 (cartesian + conversion to polar )
		 The points are expressed in [position + turret angle]
		 reference frame.
 * ARGUMENTS:    long nbpt :
 *               long *laser_val :
 * ALGORITHM:    the function takes the array "laser_val".
                 Basically,  laser val is of the type Laser[] produced
		 by the robot. 
		 A typical way to use directly with the robot would be:

		 gs();
		 Pt_array = ReadCamCart(Laser[0],Laser + 1);

 * RETURN:       p2d *
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

p2d *ReadCamCart(long nbpt, long *laser_val)

{
  long i,j,k=0;
  long Interval, Offset;
  double x,y,xi,yi;
  p2d *pt;
  double vx,vy,vxy;

  Nbpt = nbpt;

  

  Interval = 482/(Nbpt - 1);
  Offset = (482 - ((Nbpt - 1) * Interval))/2;

  pt = Pt_cree(Nbpt);
  for(j = 0; j < Nbpt; j++){

    /* testing the validity of the point 
     if not valid : Laser[j] = 65535 */
    if(laser_val[j] > 0 && laser_val[j] < 60000){

      /* retriving the original scan line in the pixel frame */
      i = Offset + j*Interval;

      /* centering the point in the pixel coordinate system */
      yi = i - 241.5;
      xi = laser_val[j] - 255.5;

      /* applying the transformation according to laser.cal */
      /* computes also the variances */

      ProjectPhyP(xi,yi,&x,&y,&vx,&vxy,&vy);
      (pt + k)->x = x;
      (pt + k)->y = y;
      
      (pt + k)->r  = sqrt(x*x + y*y);
      (pt + k)->o  = atan2(y,x);
      
      (pt + k)->vx = vx;
      (pt + k)->vxy = vxy;
      (pt + k)->vy = vy;
      
      /* transforming in polar coordinates with variances) */

      point_inc_xyro(pt+k);
      if(pt->r < MaxPtDist)
	k++;
    }
    /*********
      Otherwise,
      this point is not valid; 
    
    **********/
  }

  /******
    we put the number of valid points in the global variable "Nbpt"
  ******/
    Nbpt = k;

  return(pt);

}




/***************
 * FUNCTION:     Imgl_cart_fill
 * PURPOSE:      fills the igla structure "*ig" with the
                 the laser data contained in "*laser_val".
		 Here, the laser data is transformed into
		 physical coordinates via the calibration 
		 file "laser.cal".
 		 Before a call to this function, you have to call 
		 Read_fic_calib() (see below).

		 Points further than MaxPtDist will be discarded.
		 
		 The coordinate system is set to be the turret: 
		 ig->coosys = 'T'

		 In case you use this with the robot directly, the 
		 functions "Set_pos_image" and "Set_enc_image" allow
		 you to save encoder values in the *igla structure.
		 Here, all values are set to 0 (encoder position
		 + encoder angles + absolute position + absolute angles)

		 ig->x, ig->y = x and y absolute  of the robot
		               at perception time
		 ig->cs,ig->sn = cos and sine of encoder sterring angle
		 ig->cst, ig->snt =  cos and sine of encoder turret angle
		 ig->enc = rob_enc : x,y,st,tu
 
		 
		 All fields ->x,->y,->cs etc are intended to contain absolute
		 coordinates when different from the robots' starting point.
		 ig->enc[4] should always keep memory
		 of the encoder values at the time the perception was made.
		 
		 

 * ARGUMENTS:    igla *ig :
 *               long nbp :
 *               long *laser_val :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        ReadCamCart(), mx_init/matrix.c
 * CALLED BY:    
 ***************/


void Imgl_cart_fill(igla *ig,long nbp,long *laser_val)

{
  
  ig->tbp = ReadCamCart(nbp,laser_val);
  ig->nbp = Nbpt;
  ig->coosys = 'T';
  ig->x = 0.;
  ig->y = 0.;
  ig->cs = 1.;
  ig->sn = 0.;
  ig->cst = 1.;
  ig->snt = 0.;
  mx_init(ig->enc,4,1,0.);


}
/** change the coordinate system of ig to Sys */

void ChangeCoordSysTo(igla *ig, char Sys)
{
  Sgl *sgr;
  PDBL *l;
  double acs,asn;
  double x,y;
  p2d *pt;
  double px,py,vx,vy;

  /* nothing to do */
  if(ig->coosys == Sys)
    return;

  if(Sys == 'A'){
    if(ig->coosys == 'S')
      {
	acs = ig->cs;
	asn = ig->sn;
      }
    else if(ig->coosys == 'T')
      {
	acs = ig->cst;
	asn = ig->snt;
      }
    x = ig->x;
    y = ig->y;
  }
    
  else if (Sys == 'S'){
    if(ig->coosys == 'A')
      {
	acs = ig->cs;
	asn = -ig->sn;
	x = -ig->x;
	y = -ig->y;
      }
    else if(ig->coosys == 'T')
      {
	acs = ig->cst * ig->cs + ig->snt * ig->sn;
	asn = ig->snt * ig->cs - ig->cst * ig->sn ;
	x = 0.0;
	y = 0.0;
      }

  }

  else if(Sys == 'T'){
    if(ig->coosys == 'A')
      {
	acs = ig->cst;
	asn = -ig->snt;
	x = -ig->x;
	y = -ig->y;
      }
    else if(ig->coosys == 'S')
      {
	acs = ig->cst * ig->cs + ig->snt * ig->sn;
	asn = -ig->snt * ig->cs + ig->cst * ig->sn ;
	x = 0.0;
	y = 0.0;
      }



  }
 
  ig->coosys = Sys;
   for(l = ig->lsgl; l != NULL; l = dls(l)){

    sgr = (Sgl *)dlel(l);
    pt = &(sgr->first);
    px = pt->x;
    py = pt->y;
    pt->x = x + acs * px - asn * py;
    pt->y = y + asn * px + acs * py;
    
    pt = &(sgr->last);
    
    px = pt->x;
    py = pt->y;
    pt->x = x + acs * px - asn * py;
    pt->y = y + asn * px + acs * py;
    
    vx = sgr->vx;
    vy = sgr->vy;
    sgr->vx = vx*acs - vy*asn;
    sgr->vy = vx*asn + vy*acs;
    sgr->d = sgr->r = -sgr->vx * pt->x - sgr->vy * pt->y;
    sgr->o = atan2(sgr->vy,sgr->vx);


   }
}

/***********************************************************
  information about structures:
***********************************************************/



/***************
 * FUNCTION:     Pt_pos_x
 * PURPOSE:      returns the absolute x coordinate of the point i
                  of the image igl
 * ARGUMENTS:    igla *igl :
 *               long i :
 * ALGORITHM:    Depending on the coordinate system igl->coosys 
                 used by the image igl, this function computes the 
		 projection of igl->tbp[i] through the coordinates 
		 system defined by:
		 case "T" (turret)
		    i->x,ig->y,ig->cst,ig->snt
		 case "S" (steering)
		    i->x,ig->y,ig->cs,ig->sn
		 case "A" (absolute)
		    identity

 * RETURN:       double 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


double Pt_pos_x(igla *igl,long i)

{

  p2d *pt;

  pt = (igl->tbp + i);
  if(igl->coosys == 'T')
    return( igl->x + igl->cst * pt->x - igl->snt * pt->y);
  if(igl->coosys == 'S')
    return( igl->x + igl->cs * pt->x - igl->sn * pt->y);
  return(pt->x);
}



/***************
 * FUNCTION:     Pt_pos_y
 * PURPOSE:      returns the absolute y coordinate of the point i
                  of the image igl
 * ARGUMENTS:    igla *igl :
 *               long i :
 * ALGORITHM:    Depending on the coordinate system igl->coosys 
                 used by the image igl, this function computes the 
		 projection of igl->tbp[i] through the coordinates 
		 system defined by:
		 case "T" (turret)
		    i->x,ig->y,ig->cst,ig->snt
		 case "S" (steering)
		    i->x,ig->y,ig->cs,ig->sn
		 case "A" (absolute)
		    identity

 * RETURN:       double 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

double Pt_pos_y(igla *igl,long i)

{

  p2d *pt;

  pt = (igl->tbp + i);
  if(igl->coosys == 'T')
    return( igl->y + igl->snt * pt->x + igl->cst * pt->y);
  if(igl->coosys == 'S')
    return(igl->y + igl->sn * pt->x + igl->cs * pt->y);
  return(pt->y);
}




/***************
 * FUNCTION:     Pt_pos_y
 * PURPOSE:      compute the absolute  coordinates of the point i
                  of the image igl. p[0] = x, p[1] = y.
 * ARGUMENTS:    igla *igl :
 *               long i :
                 double *p:
 * ALGORITHM:    Depending on the coordinate system igl->coosys 
                 used by the image igl, this function computes the 
		 projection of igl->tbp[i] through the coordinates 
		 system defined by:
		 case "T" (turret)
		    i->x,ig->y,ig->cst,ig->snt
		 case "S" (steering)
		    i->x,ig->y,ig->cs,ig->sn
		 case "A" (absolute)
		    identity

 * RETURN:       double 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void Pt_pos(igla *igl,p2d *pt,double *p)

{

  if(igl->coosys == 'T'){
    p[0] = igl->x + igl->cst * pt->x - igl->snt * pt->y;
    p[1] = igl->y + igl->snt * pt->x + igl->cst * pt->y;
    return;
  }
  if(igl->coosys == 'S'){
    p[0] = igl->x + igl->cs * pt->x - igl->sn * pt->y;
    p[1] = igl->y + igl->sn * pt->x + igl->cs * pt->y;
    return;
  }
  p[0] = pt->x;
  p[1] = pt->y;

  return;
    
}

/***************
 * FUNCTION:     Sg_pos
 * PURPOSE:      compute the absolute  coordinates of the segment *sg
                  of the image igl. 
		  p[0] = x of first point (sg->first->x)
		  p[1] = y of first point (sg->first->y)
		  p[2] = x of last point (sg->last->x)
		  p[3] = y of last point (sg->last->y)
		  
 * ARGUMENTS:    igla *igl :
 *               long i :
                 double *p:
 * ALGORITHM:    Depending on the coordinate system igl->coosys 
                 used by the image igl, this function computes the 
		 projection of igl->tbp[i] through the coordinates 
		 system defined by:
		 case "T" (turret)
		    i->x,ig->y,ig->cst,ig->snt
		 case "S" (steering)
		    i->x,ig->y,ig->cs,ig->sn
		 case "A" (absolute)
		    identity

 * RETURN:       double 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


void Sg_pos(igla *igl,Sgl *sg,double *p)

{

  p2d *pt;

  

  if(igl->coosys == 'T')
    {
      pt = &(sg->first);
      p[0] = igl->x + igl->cst * pt->x - igl->snt * pt->y;
      p[1] = igl->y + igl->snt * pt->x + igl->cst * pt->y;
      pt = &(sg->last);
      p[2] = igl->x + igl->cst * pt->x - igl->snt * pt->y;
      p[3] = igl->y + igl->snt * pt->x + igl->cst * pt->y;
      return;
    }
  
  if(igl->coosys == 'S')
    {
       pt = &(sg->first);
       p[0] = igl->x + igl->cs * pt->x - igl->sn * pt->y;
       p[1] = igl->y + igl->sn * pt->x + igl->cs * pt->y;
       pt = &(sg->last);
       p[2] = igl->x + igl->cs * pt->x - igl->sn * pt->y;
       p[3] = igl->y + igl->sn * pt->x + igl->cs * pt->y;
       return;
    }
  pt = &(sg->first);
  p[0] = pt->x;
  p[1] = pt->y;
  pt = &(sg->last);
  p[2] = pt->x;
  p[3] = pt->y;
  
  return;
       
}




/***************
 * FUNCTION:     Proj_sg
 * PURPOSE:      computes the projection of sg in the coordinate system defined
                 in igl:

		 case "T" (turret)
		    i->x,ig->y,ig->cst,ig->snt
		 case "S" (steering)
		    i->x,ig->y,ig->cs,ig->sn
		 case "A" (absolute)
		    identity
		 Computes the polar representation of the new segment.

                 returns the projected segment;
		 
                 Warning: creates the segment result, you have to free it.
 * ARGUMENTS:    igla *igl :
 *               Sgl *sg :
 * ALGORITHM:    
 * RETURN:       Sgl *
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

Sgl *Proj_sg(igla *igl,Sgl *sg)


{

  p2d *pt,*pi;
  Sgl *sgr;

  if((sgr = Sgl_cree()) == NULL)
    return(NULL);

  *sgr = *sg;
 
  pt = &(sg->first);
  pi = &(sgr->first);

  if(igl->coosys == 'T'){
    pi->x = igl->x + igl->cst * pt->x - igl->snt * pt->y;
    pi->y = igl->y + igl->snt * pt->x + igl->cst * pt->y;
    
    pt = &(sg->last);
    pi = &(sgr->last);
    
    pi->x = igl->x + igl->cst * pt->x - igl->snt * pt->y;
    pi->y = igl->y + igl->snt * pt->x + igl->cst * pt->y;
    
    sgr->vx = sg->vx*igl->cst - sg->vy*igl->snt;
    sgr->vy = sg->vx*igl->snt + sg->vy*igl->cst;
    sgr->d = sgr->r = -sgr->vx * pi->x - sgr->vy * pi->y;
    sgr->o = atan2(sgr->vy,sgr->vx);

    
    
    return(sgr);
  }
  if(igl->coosys == 'S'){
    pi->x = igl->x + igl->cs * pt->x - igl->sn * pt->y;
    pi->y = igl->y + igl->sn * pt->x + igl->cs * pt->y;
    
    pt = &(sg->last);
    pi = &(sgr->last);
    
    pi->x = igl->x + igl->cs * pt->x - igl->sn * pt->y;
    pi->y = igl->y + igl->sn * pt->x + igl->cs * pt->y;
    
    sgr->vx = sg->vx*igl->cs - sg->vy*igl->sn;
    sgr->vy = sg->vx*igl->sn + sg->vy*igl->cs;
    sgr->d = sgr->r= -sgr->vx * pi->x - sgr->vy * pi->y;
    sgr->o = atan2(sgr->vy,sgr->vx);
    return(sgr);
  }
  pi->x = pt->x;
  pi->y = pt->y;
  
  pt = &(sg->last);
  pi = &(sgr->last);
  
  pi->x = pt->x;
  pi->y = pt->y;
  
  sgr->vx = sg->vx;
  sgr->vy = sg->vy;
  sgr->d = sgr->r = sg->d;
  sgr->o = sg->o;

  return(sgr);
  
}

/***************
 * FUNCTION:     Proj_ima_s
 * PURPOSE:      projects segments in *igl in the absolute 
                 coordinate system using *igl coordinate system.
		 Copies the resulting segments in image *ig
		 ig coordinate system is 0,0,0,0.
		 ig->enc = igl->enc

		 ig's segments are allocated in this function
		 ig's structure has to be allocated before
		 calling this function.

		 points in igl->tbp (if any) are not projected
		 variance matrices are merely copied without
		 transformations

		 Returns 1 on success
		 Returns 0 if memory allocation pbs.

 * ARGUMENTS:    igla *igl :
 *               igla *ig :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        mx_copy/matrix.c, Imgl_free
 * CALLED BY:    
 ***************/

long Proj_ima_s(igla *igl,igla *ig)

{

  PDBL *l,*l2;
  Sgl *sg,*sm;


  

  ig->x = 0;
  ig->y = 0;
  ig->cs = 1.;
  ig->sn = 0.;
  ig->cst = 1;
  ig->snt = 0;
  ig->coosys = 'A';
  mx_copy(igl->enc,ig->enc,4,1);

  ig->lsgl = NULL;
  ig->nsg = igl->nsg;

  for(l = igl->lsgl; l != NULL; l = dls(l)){

    sm = (Sgl *)dlel(l);
    
    if((sg = Proj_sg(igl,sm)) == NULL)
      {
	Imgl_free(ig);
	return(0);
      }

    sg->vr = sm->vr; sg->vo = sm->vo; sg->vro = sm->vro;

    if(ig->lsgl == NULL){

      ig->lsgl = dlbcons(NULL,(char *)sg);
      l2 = ig->lsgl;
    }
    else
      l2 = dlbcons(l2,(char *)sg);
  }
  return(1);
}


/***************
 * FUNCTION:     Set_pos_image
 * PURPOSE:      assigns a new absolute coordinate system to 
                 the image *igl.
		
		 th and tu are in radian
		 this new coordinates system is set in the fields
		 igl->x,igl->y : absolute x and y
		 igl->cs, igl->sn : cos and sin of steering angle
		 igl->cst,igl->snt : cos and sin of turret angle

 * ARGUMENTS:    igla *igl :
 *               double x :
 *               double y :
 *               double th :
 *               double tu :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


void Set_pos_image(igla *igl,double x,double y,double th,double tu)


{

  igl->x = x;
  igl->y = y;
  igl->cs = cos(th);
  igl->sn = sin(th);
  igl->cst = cos(tu);
  igl->snt = sin(tu);

}


/***************
 * FUNCTION:     Set_enc_image
 * PURPOSE:      saves the encoder values in *igl
                 *encoders values are supposed to
		 be longs integers in 10th of inches
		 and 10th or degrees.
		 They are converted in inches and radian
		 igl->enc[0] = x
		 igl->enc[1] = y
		 igl->enc[2] = steering angle
		 igl->enc[3] = turret angle
 * ARGUMENTS:    igla *igl :
 *               long *encoders :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void Set_enc_image(igla *igl,long *encoders)
{
  igl->enc[0] = encoders[0]/10.;
  igl->enc[1] = encoders[1]/10.;
  igl->enc[2] = M_PI*encoders[2]/1800.;
  igl->enc[3] = M_PI*encoders[3]/1800.;
}
  

/***************
 * FUNCTION:     Imgl_create_copy
 * PURPOSE:      returns an indexed copy
                 of *igm.
		 The main purpose is to be able to transform,
		 even truncate, the segments in the resulting image
		 while keeping memory about the original segment in
		 *igm.
		 
 * ARGUMENTS:    igla *igm :
 * ALGORITHM:    all segment in igm->lsgl are copied onto
                 igr->lsgl
		 all segments in igr->lsgl contain the pointer
		 on the corresponding *igm segment in their 
		 ->sfirst field
		 the field igr->usr becomes an array representation
		 of the list igr->lsgl.
		 This is mainly done for aspect computation when we want
		 to know from where are coming the local visible segments.
 * RETURN:       igla *
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

igla *Imgl_create_copy(igla *igm)
{
  igla *igr;
  Sgl *s;
  long i;
  Sgl **c;
  PDBL *l,*la = NULL;

  igr = Imgl_cree();
  c  = (Sgl **)malloc(igm->nsg * sizeof(Sgl *));
  igr->usr = (char *)c;

  for(l = igm->lsgl; l != NULL; l = dls(l)){
    s = Sgl_cree();

    /* FIFO statcking */

    la = dlbcons(la,s);
    if(igr->lsgl == NULL)
      igr->lsgl = la;

    /** added 05/17/94:
      I copy the pointer on the global map
      in the ->sfirst field (hacky, isn't it ?)
      for retreiving it from the local
      segment
    **/

    s->sfirst = (Sgl *)(dlel(l));
  }
  
  i = 0;
  for(l = igr->lsgl; l != NULL; l = dls(l)){
    c[i] = (Sgl *)dlel(l);
    i++;
  }
  igr->nsg = igm->nsg;
  
  return(igr);
}


/**************************************************************
  Transformations according to the calibration
**************************************************************/



/***************
 * FUNCTION:     read_laser_cal
 * PURPOSE:      reads the file specified in the argument
                 sets the parameters CamParam[8] used for 
		 transforming pixel points into physical 
		 coordinates. 
		 sets the variance on the parameters in
		 VarParam[]
		 WARNING: the file must have been produced by
		 the calibration program.
 * ARGUMENTS:    char *file :
 * ALGORITHM:    
 * RETURN:       long: = 0 if file ok
                       = -1 if pb with the file 
 * SIDE EFFECT:  fills up CamParam[] and VarParam[]
                 changes the static variable "File"
 * CALLS:        Read_init_calib()
 * CALLED BY:    
 ***************/

long read_laser_cal(char *file)
{

  *File = *file;
  return(Read_init_calib());
}


/***************
 * FUNCTION:     Read_fic_calib
 * PURPOSE:      interactively asks for a calibration file
                 name and calls Read_init_calib() as above.
		 WARNING: the file must have been produced by
		 the calibration program.
 * ARGUMENTS:    None
 * ALGORITHM:    
 * RETURN:       long = 0 if file ok
                      = -1 if pb with the file 
 * SIDE EFFECT:  fills up CamParam[] and VarParam[]
                 changes the static variable "File"
 * CALLS:        Read_init_calib()
 * CALLED BY:    
 ***************/

long Read_fic_calib(void)
{
  printf("calibration file name ?\n");
  scanf("%s",File);
  return(Read_init_calib());
}




/***************
 * FUNCTION:     Read_init_calib
 * PURPOSE:      read the file "File";
                 by default File is "laser.cal";
		 If you want to use another file, 
		 call read_laser_cal() (linked
		 version) or Read_fic_calib() (interactive 
		 version)
		 WARNING: the file must have been produced by
		 the calibration program.
 * ARGUMENTS:    None
 * ALGORITHM:    
 * RETURN:       long = 0 if file ok
                      = -1 if pb with the file 
 * SIDE EFFECT:  fills up CamParam[] and VarParam[]
                 changes the static variable "File"
 * CALLS:       
 * CALLED BY:    
 ***************/

long Read_init_calib(void)

{

  long i;

  if((ptr = fopen(File,"r")) == NULL){
    printf("Can't open file %s \n",File);
    return(-1);
  }

  fscanf(ptr,"%lf %lf %lf %lf",
	 &CamParam[0],&CamParam[1],&CamParam[2],&CamParam[3]);
  fscanf(ptr,"%lf %lf %lf %lf",
	 &CamParam[4],&CamParam[5],&CamParam[6],&CamParam[7]);

  fscanf(ptr,"%lf %lf",
	 &TransRobot[0],&TransRobot[1]);

  for(i = 0; i < 36; i += 6){
    fscanf(ptr,"%lf %lf %lf %lf %lf %lf",
	   VarParam + i,VarParam+i+1,VarParam+i+2,
	   VarParam+i+3,VarParam+i+4,VarParam+i+5);
  }

  fclose(ptr);

  return(0);
}


/***********************
Transformation function according to the calibration ,
   The resulting points are in the turret frame
   (i.e, : x,y and tu)

   yi = scanline
   xi = laser_val[yi]

   *x and *y are the resulting coordinates
***********************/


/***************
 * FUNCTION:     ProjectPhy
 * PURPOSE:      Transformation function according to the calibration ,
                 The resulting points are in the turret frame
                 (i.e, : x,y and tu)

		 yi = scanline
		 xi = laser_val[yi] in PIXEL

		 *x and *y are the resulting coordinates
		 WARNING: a call to Read_init_calib must have been
		 made for this function to compute correct values.
		 
 * ARGUMENTS:    double xi :
 *               double yi :
 *               double *x :
 *               double *y :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void ProjectPhy(double xi,double yi,double *x,double *y)


{
  double den;

  den = (CamParam[0]*xi + CamParam[1]*yi + 1);

  *x = (CamParam[2]*xi + CamParam[3]*yi + CamParam[4])/den + TransRobot[0];
  *y = (CamParam[5]*xi + CamParam[6]*yi + CamParam[7])/den + TransRobot[1];

}


/***************
 * FUNCTION:     pixtorange
 * PURPOSE:      Transformation function according to the calibration ,
                 The resulting points are in the turret frame
                 (i.e, : x,y and tu)

		 yi = scanline
		 xi = laser_val[yi] in PIXEL

		 *x and *y are the resulting coordinates
		 Same as above but xi and yi are integers and 
		 x and y represents tenth of the basic unit of
		 the calibration.

		 WARNING: a call to Read_init_calib must have been
		 made for this function to compute correct values.
		 
 * ARGUMENTS:    long xi :
 *               long yi :
 *               long *x :
 *               long *y :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


void pixtorange(long xi,long yi,long *x,long *y)


{
  double den;

  den = (CamParam[0]*xi + CamParam[1]*yi + 1);

  *x = 10 * ( (CamParam[2]*xi + CamParam[3]*yi + CamParam[4])/den + TransRobot[0]);
  *y = 10 * ( (CamParam[5]*xi + CamParam[6]*yi + CamParam[7])/den + TransRobot[1]);

}


/*************
  same thing as ProjectPhy but computes also
  the variance on x and y in vx,vxy and vy
*************/


/***************
 * FUNCTION:     ProjectPhyP
 * PURPOSE:      Transformation function according to the calibration 
                 Physical points + Variance.

                 The resulting points are in the turret frame
                 (i.e, : x,y and tu)

		 yi = scanline
		 xi = laser_val[yi] in PIXEL

		 *x and *y are the resulting coordinates
		 vx,vxy and vy represents the covariance
		 matrix :

		 [vx vxy]
		 [vxy vy]

		 WARNING: a call to Read_init_calib must have been
		 made previously for this function to compute correct values.
		 
 * ARGUMENTS:    double xi :
 *               double yi :
 *               double *x :
 *               double *y :
 *               double *vx :
 *               double *vxy :
 *               double *vy :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:   mxfs_mmulttr()/matrix.c     
 * CALLED BY:    
 ***************/


void ProjectPhyP(double xi,double yi,double *x,double *y,double *vx,
		    double *vxy,double *vy)


{
  double den,den2,num1,num2;

  den = (CamParam[0]*xi + CamParam[1]*yi + 1);
  den2 = den*den;
  num1 = (CamParam[2]*xi + CamParam[3]*yi + CamParam[4]);
  num2 = (CamParam[5]*xi + CamParam[6]*yi + CamParam[7]);
  *x = num1/den + TransRobot[0];
  *y = num2/den + TransRobot[1];

  /* Jpar = jacobian of {num1/den,num2/den} with respect to CamParam */

  Jpar[0] = -xi*num1/den2; Jpar[1] = -yi*num1/den2; Jpar[2] = xi/den;
  Jpar[3] = yi/den;
  Jpar[4] = 1./den; Jpar[5] = 0.; Jpar[6] = 0.; Jpar[7] = 0.;

  Jpar[8] = -xi*num2/den2; Jpar[9] = -yi*num2/den2; Jpar[10] = 0.;
  Jpar[11] = 0.;
  Jpar[12] = 0.; Jpar[13] = Jpar[2]; Jpar[14] = Jpar[3]; Jpar[15] = 1./den;

  /* error contribution due to the error on the paramters
     er = Jpar Epar => Vpar = Jpar.VarParam.transpose(Jpar) */

  mxfs_mmulttr(Jpar,VarParam,Vxpp,2,8,8,8);

  /* Jxpi = jacobian with respect to xi and yi */

  Jxi[0] = (CamParam[2]*den - CamParam[0]*num1)/den2;
  Jxi[1] = (CamParam[3]*den - CamParam[1]*num1)/den2;
  Jxi[2] = (CamParam[5]*den - CamParam[0]*num2)/den2;
  Jxi[3] = (CamParam[6]*den - CamParam[1]*num2)/den2;

  Vxpi[0] = Jxi[0]*Jxi[0]*vpix_x + Jxi[1]*Jxi[1]*vpix_y;
  Vxpi[1] = Jxi[0]*Jxi[2]*vpix_x + Jxi[1]*Jxi[3]*vpix_y;
  Vxpi[2] = Jxi[2]*Jxi[2]*vpix_x + Jxi[3]*Jxi[3]*vpix_y;

  *vx = Vxpp[0] + Vxpi[0] + 0.2;
  *vxy = Vxpp[1] + Vxpi[1];
  *vy = Vxpp[2] + Vxpi[2] + 0.2;

  
}















