/* 
 *   Functions to deal with uncertainty of the laser data
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef VXWORKS
#define M_PI 3.141592653589793
#endif
#include "mat.h"
#include "gl_prot.h"

 
#include "kal_prot.h"

#ifndef max
#define max(x,y) (x>y ? x : y)
#endif
#ifndef min
#define min(x,y) (x>y ? y : x)
#endif

static double vs[3],vp[3];

static double Jac[9],jac1[4],jac2[4];
static double vaux1[3],vaux2[3];



/**************************************

  This file deals with uncertainty computation
  and more precisely with variance matrix computation.
  
  There are two cases:
  Knowing the variance on an object x
  we want to compute the variance on 
  
  y = f(x)

  If there is an error on x, we do a first 
  order aproxmiation:

  y + e_y = f(x) + Jacobian(f)/x * e_x

  leading to 

  e_y = Jacobian(f)/x * e_x

  Then, 

  V_y = E[e_y e_y^t] = J(f)/x V_x J(f)^t

  Knowing a variance on x0
  it is often easier to compute the variance
  on object y0 from an implicit relation
  
  f(x0,y0) = 0

  That is usually possible only if the 
  implicit theorem condition is satisfied
  around the point y0, that is, if

  Jacobian(f)/y = J_y is regular at y=y0

  In this case, we get 

  J(f)/y e_y + J(f)/x e_x = 0

  leading to 

  V_y = (J(f)/y)^(-1) J_x V_x J(f)_x^t ((J(f)/y)^(-1))^t

  Note: this form is summarized by V_y = A V A^t
        V being a symmetrical matrix.
	This is implemented in the matrix library by:

  mxfs_mmulttr(A,V,V_y,...) (see readme.user in libmatrix directory)

  ***********************************************/

  




/*
 *  Transform of the uncertainty from   cartesian to polar
 */


/**** 

*****/

/***************
 * FUNCTION:     impt_inc_xyro
 * PURPOSE:      Transform of the uncertainty of an entire image 
                 of points from cartesian to polar. 
		 The transform from pixel points to cartesian 
		 coordinates computes the uncertainty in cartesian
		 coordinates. This function computes the uncertainty
		 in polar coordinates
 * ARGUMENTS:    igla *ig :
 * ALGORITHM:    calls point_inc_xyro for all points in *ig
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        point_inc_xyro()
 * CALLED BY:    
 ***************/

void impt_inc_xyro(igla *ig)
    
{
  
  
  int i,npt;
  p2d *pt;

  
  pt = ig->tbp;
  npt = ig->nbp;
  
  for(i = 0; i < npt; i++){
    point_inc_xyro(pt+i);

  }

}

/**************
  Transform for a single point
***************/



/***************
 * FUNCTION:     point_inc_xyro
 * PURPOSE:      transforms the variance matrix from cartesian to
                 polar and write it in the fields
		 pt->vr variance on rho
		 pt->vo variance on theta
		 ptr->vro covariance rho-theta
 * ARGUMENTS:    p2d *pt :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:     mxfs_mmulttr()/matrix.c   
 * CALLED BY:    
 ***************/

void point_inc_xyro(p2d *pt)


{

  /* computing the jacobians of the transform function */
  Jac[0] = pt->x /pt->r;
  Jac[1] = pt->y/pt->r;
  Jac[2] = -pt->y/(pt->r*pt->r);
  Jac[3] = pt->x/(pt->r*pt->r);

  /* putting variance on cartesian under matrix form */
  vs[0] = pt->vx; vs[1] = pt->vxy; vs[2] = pt->vy;

  /* computing vp */
  mxfs_mmulttr(Jac,vs,vp,2,2,2,2);

  /* copying in the point structure */
  pt->vr = vp[0]; pt->vo = vp[2]; pt->vro = vp[1];
  

}

/***************
 * FUNCTION:     point_inc_roxy
 * PURPOSE:      transforms the variance matrix of the point pt stored in
                 pt->vr, pt->vo and pt->vro
                 from polar to cartesian and write it in the fields
		 pt->vx variance on x
		 pt->vy variance on y
		 ptr->vxy covariance x-y
 * ARGUMENTS:    p2d *pt :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:     mxfs_mmulttr()/matrix.c   
 * CALLED BY:    
  ***************/

void point_inc_roxy(p2d *pt)


{
  static double v_pol[3];



  v_pol[0] = pt->vr;
  v_pol[1] = pt->vro;
  v_pol[2] = pt->vo;
  
  /* computing the jacobians of the transform function */
  Jac[0] = cos(pt->o);
  Jac[1] = -pt->r * sin(pt->o);
  Jac[2] = sin(pt->o);
  Jac[3] = pt->r * cos(pt->o);

  
  /* computing vp */
  mxfs_mmulttr(Jac,v_pol,vp,2,2,2,2);

  /* copying in the point structure */
  pt->vx = vp[0]; pt->vy = vp[2]; pt->vxy = vp[1];
  

}

/*
 *  
 *
 *  The grouping itself has been done by segmentation() in lasegm.c which
 *  eventualy calls this function at the very end.
 *  The list of groups is in fact a list of segments. The fields
 *  sg->iend and sg->ifirst are the indices of the limits of each group in the
 *  array igl->tbp
 */


/***************
 * FUNCTION:     las_segm_unc
 * PURPOSE:      Kalman's estimate of the image of segments igl

                 Warning, the points to be segmented are supposed to be in
		 igl->tbp. You should have used Imgl_cat_fill or
		 Imgl_cam_fill in las_struct.c before calling this
		 function.

		 The point segmentation could have been made by segmentation()
		 in lasegm.c which eventually calls this function at the
		 very end. However, one can use its own segmentation
		 function as long as the points are in igl->tbp and that,
		 for each segment in igl->lsgl, the field sg->ifirst
		 (repectively sg->ilast) is the index of the first (rep. last)
		 point in the array igl->tbp.

		 For each segment, we initilize value plus variance
		 from its two extrmity points using initdt2() (see below)
		 Then, each remaning points are used to refine the segment
		 estimation by applying "fusion()" in kalplus.c.
		 The observation function "dist()" and its jacobians are 
		 declared below.

 * ARGUMENTS:    igla *igl :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        initdt2(), fusion()/kalplus.c
 * CALLED BY:    
 ***************/

void las_segm_unc(igla *igl)

{

  PDBL *list; /* list of segments of igl : igl->lsgl */
  p2d *pt;
  struct dt2d *sgl;
  int i,last,first;
  double sg[2],vsg[3],p[2],vp[3];


  for (list = igl->lsgl;list != NULL;list = dls(list))
    {
      sgl = (Sgl *)(dlel(list));

      /* last = min((prev-1),sgl->ilast); */
      last = sgl->ilast;
      first = sgl->ifirst;

      /* Particular case of a 2-point segment */
      if (first == last) last++;

      
      
      /* 
	 if the last points are very close, their variance is very
	 small and error on the least square segmentation can becomes
	 very devastating. In this case, we don't use the last three points 
	 at the end of the segment
      */
      /* compiling option in case we have a laser with constant uncertainty
	 over its perception range like the SICK 
	 */


#ifndef SICK
/*       if(igl->tbp[last].r <= 40.) */
/* 	last = last-3; */
#endif 

      /* Initialization with the 2 extemity points  */
      initdt2(igl->tbp + first,igl->tbp + last, sgl);

      /* Transforming into an array representation for Kalman Filtering */

      sg[0] = sgl->r;  sg[1] = sgl->o;
      vsg[0] = sgl->vr; vsg[1] = sgl->vro; vsg[2] = sgl->vo;

      /** loop over the remaining points */
      for (i=last-1;i>first;i--)

	{
	  pt = igl->tbp + i;
	  p[0] = pt->r;  p[1] = pt->o;
	  if(p[0] != 0. || p[1] != 0.){
	     vp[0] = pt->vr; vp[1] = pt->vro; vp[2] = pt->vo;
	     fusion (dist,ddistp,ddistsg,p,vp,
		     sg,vsg,NULL,sg,vsg,1,2,2);
	   }
	}


      /* filling up the segment structure */
      sgl->r = sg[0]; sgl->o = sg[1];
      sgl->vr = vsg[0]; sgl->vro = vsg[1]; sgl->vo = vsg[2];
      sgl->vx = cos(sgl->o); sgl->vy = sin(sgl->o);
      sgl->d = sgl->r;

      /* prev = first; */

    }
}



/***************
 * FUNCTION:     intersg_unc
 * PURPOSE:      computes the uncertain intersection *ptp + its variance matrix
                 between 2 uncertain lines *sg1 and *sg2
		 		 
 * ARGUMENTS:    Sgl *sg1 :
 *               Sgl *sg2 :
 *               p2d *ptp :
 * ALGORITHM:    
 * RETURN:       int: status on the intersection, = 0 if segments
                      are parallel, 1 otherwise.
 * SIDE EFFECT:  
 * CALLS:        intersg2_unc()
 * CALLED BY:    
 ***************/

int intersg_unc(Sgl *sg1,Sgl *sg2,p2d *ptp)
  

{



  /* computing the intersection (non-proba) */

  if(intersg(sg1,sg2,ptp) == 0)
    return(0); /* segments are parralel */

  /* computing the uncertainty on ptp */
  intersg2_unc(sg1,sg2,ptp);

  return(1);
}



/***************
 * FUNCTION:     intersg2_unc
 * PURPOSE:      computes the uncertainty of the intersection *ptp 
                 between 2 uncertain lines *sg1 and *sg2. The intersection
		 itself is supposed already computed.
		 		 
 * ARGUMENTS:    Sgl *sg1 :
 *               Sgl *sg2 :
 *               p2d *ptp :
 * ALGORITHM:    
 * RETURN:       int: status on the intersection, = 0 if segments
                      are parallel, 1 otherwise.
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

int intersg2_unc(Sgl *sg1,Sgl *sg2,p2d *ptp)
    
{


  double vx1,vy1,vx2,vy2;
  double cs,sn,fac1,fac2;
  

  /* copying variances under matrix format */ 
  vs[0] = sg1->vr; vs[1] = sg1->vro; vs[2] = sg1->vo;
  vp[0] = sg2->vr; vp[1] = sg2->vro; vp[2] = sg2->vo;

  /* 
     computing the Jacobians of
     the equation ptp = Intersection(sg1,sg2)
     jac1 is about sg1, jac2 is about sg2
  */

  vx1 = sg1->vx ; vy1 = sg1->vy;
  vx2 = sg2->vx; vy2 = sg2->vy;

  sn = -vx2*vy1 + vx1*vy2;
  cs = vx1*vx2 + vy1*vy2;

  fac1 = (-sg2->r + sg1->r*cs)/(sn*sn);
  fac2 = (-sg1->r + sg2->r*cs)/(sn*sn);

  jac1[0] = -vy2/sn; jac1[1] = -vy2*fac1;
  jac1[2] = vx2/sn; jac1[3] = vx2*fac1;

  jac2[0] = vy1/sn; jac2[1] = -vy1*fac2;
  jac2[2] = -vx1/sn; jac2[3] = vx1*fac2;


  /* Multiplying by the Jacobians */

  mxfs_mmulttr (jac1, vs, vaux1, 2,2,2,2);
  mxfs_mmulttr (jac2, vp, vaux2, 2,2,2,2);

  /* adding the respectives uncertainties */

  mxs_plus(vaux1,vaux2,vaux1,2,2);

  /* filling the point structure */

  ptp->vx  = vaux1[0];
  ptp->vxy = vaux1[1];
  ptp->vy  = vaux1[2];

  return(1);

}



/*********
  initialize a line with its variance from 
  p1 and p2 with their variance.

  Cartesian coordinates version.

  Convention: ->o is the angle (radian) of the opposite normal 
              ->d is the orthogonal distance to the origin
	      ->vx = -cos(o)
	      ->vy = -sin(o)
	      vx,vy, is the normal going to the right when you walk from 
	      p1 to p2

  Looks better than the same thing in polar coodinates.
  Remains to be explained ....
***********/
     

/***************
 * FUNCTION:     initdt2
 * PURPOSE:      initialize a line with its variance from 
                 p1 and p2 with their variance.

		 Convention: ->o is the angle (radian) of the opposite normal 
		             ->d is the orthogonal distance to the origin
			     ->vx = -cos(o)
			     ->vy = -sin(o)
			     vx,vy, is the normal going to the right when
			     you walk from  p1 to p2
		 The line *dt is filled with the result
		 The function returns 0 if p1 and p2 are identical.
		 Returns 1 otherwise.

 * ARGUMENTS:    p2d *p1 :
 *               p2d *p2 :
 *               Sgl *dt :
 * ALGORITHM:    
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

int initdt2 (p2d *p1,p2d *p2,Sgl *dt)

{
  double teta,d,delta;
  double x1,y1,x2,y2,X,Y;
  double cs,sn;
  static double v[10] = {0.,0.,0.,0.,0.,0.,0.,0.,0.,0.};
  /* if using mapple 
  double t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19;
  */
  x1 = p1->x;
  y1 = p1->y; 
  x2 = p2->x; 
  y2 = p2->y;
  X = x2 - x1; Y = y2 - y1;

  delta = X*X + Y*Y;

  if (delta == 0) return(1);

  teta = atan2(X,-Y);
  cs = cos(teta); sn = sin(teta);

  
  d = -cs*x1 -sn*y1;
  dt->r = dt->d = d;
  dt->o = teta;

  Jac[4] = Y/delta;
  Jac[5] = -X/delta;
  Jac[6] = -Y/delta;
  Jac[7] = X/delta;
  
  /* according to Mapple .... */
  /*
   t1 = x2-x1;
   t2 = -y2+y1;
   t3 = atan2(t1,t2);
   t5 = 1/t2;
   t7 = t1*t1;
   t8 = t2*t2;
   t12 = 1/(1.0+t7/t8);
   t15 = cos(t3);
   Jac[0] = -sin(t3)*t5*t12*x1-t15+t15*t5*t12*y1;
   
   t1 = x2-x1;
   t2 = -y2+y1;
   t3 = atan2(t1,t2);
   t4 = sin(t3);
   t6 = t2*t2;
   t7 = 1/t6;
   t8 = t1*t1;
   t12 = t7/(1.0+t8*t7);
   
   Jac[1] = -t4*t1*t12*x1+cos(t3)*t1*t12*y1-t4;

   t1 = x2-x1;
   t2 = -y2+y1;
   t3 = atan2(t1,t2);
   t5 = 1/t2;
   t7 = t1*t1;
   t8 = t2*t2;
   t12 = 1/(1.0+t7/t8);
   Jac[2] = sin(t3)*t5*t12*x1-cos(t3)*t5*t12*y1;
   
   t1 = x2-x1;
   t2 = -y2+y1;
   t3 = atan2(t1,t2);
   t6 = t2*t2;
   t7 = 1/t6;
   t8 = t1*t1;
   t12 = t7/(1.0+t8*t7);
   Jac[3] = sin(t3)*t1*t12*x1-cos(t3)*t1*t12*y1;
   */
  
  x2 = sn*x1 - cs*y1;

    
  Jac[0] =  x2*Jac[4] - cs;
  Jac[1] =  x2*Jac[5] - sn;
  Jac[2] =  x2*Jac[6];
  Jac[3] =  x2*Jac[7];
      

  /* sara 2/11/3 : bug : left part was not modifiable ! */
/*   Smelt(v,0,0) = p1->vx; Smelt(v,1,0) = p1->vxy; Smelt(v,1,1) = p1->vy; */
/*   Smelt(v,2,2) = p2->vx; Smelt(v,2,3) = p2->vxy; Smelt(v,3,3) = p2->vy; */

  v[Smelt_ind(0,0)] = p1->vx; 
  v[Smelt_ind(1,0)] = p1->vxy; 
  v[Smelt_ind(1,1)] = p1->vy;
  v[Smelt_ind(2,2)] = p2->vx; 
  v[Smelt_ind(2,3)] = p2->vxy; 
  v[Smelt_ind(3,3)] = p2->vy;

  mxfs_mmulttr(Jac,v,vs,2,4,4,4);

  dt->vr = vs[0]; dt->vro = vs[1]; dt->vo = vs[2];
  return(0);
}




/* Kalman's Observation Functions passed to fusion()/kalplus.c */

/* line equation in polar coordinates 
   
   ro(pt)*cos(phi(pt) - phi(sg)) - dist2origin(sg) = 0

*/


/***************
 * FUNCTION:     dist
 * PURPOSE:      puts in d the result of the line equation
                 ro(pt)*cos(phi(pt) - phi(sg)) - dist2origin(sg)
		 Note: d is in fact the distance from pt to sg.
		 This function is intended to be passed as a parameter
		 to the generic kalman function "fusion()".
 * ARGUMENTS:    double *pt :
 *               double *sg :
 *               double *d :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void dist(double *pt,double *sg,double *d)

{
  *d = -pt[0]*cos(pt[1]-sg[1]) - sg[0];
}


/***************
 * FUNCTION:     ddistp
 * PURPOSE:      jacobian of "dist" around pt
                 Function going together with dist
		 when used by fusion()
 * ARGUMENTS:    double *pt :
 *               double *sg :
 *               double *dp :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void ddistp(double *pt,double *sg,double *dp)

{
  dp[0] = -cos(pt[1]-sg[1]);
  dp[1] = pt[0]*sin(pt[1]-sg[1]);
}


/***************
 * FUNCTION:     ddistsg
 * PURPOSE:      jacobian of "dist" around sg
                 Function going together with dist
		 when used by fusion()
 * ARGUMENTS:    double *pt :
 *               double *sg :
 *               double *dsg :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


void ddistsg(double *pt,double *sg,double *dsg)

{
  dsg[0] = -1.0;
  dsg[1] = -pt[0]*sin(pt[1]-sg[1]);
}




/********
  special Kalman identity functions when the x[1] is an angle.
********/


/***************
 * FUNCTION:     idta
 * PURPOSE:      f = x - z in 2D where x[1] and z[1] 
                 are angles and need to be normalized
 * ARGUMENTS:    double *z :
 *               double *x :
 *               double *f :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        norm_ang()/lasutil.c
 * CALLED BY:    
 ***************/

void idta(double *z,double *x,double *f)
{
  f[0] = x[0] - z[0];
  f[1] = norm_ang(x[1] - z[1],-M_PI);
  

}


/***************
 * FUNCTION:     didtax
 * PURPOSE:      jacobian of idta around x
 * ARGUMENTS:    double *z :
 *               double *x :
 *               double *r :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void didtax(double *z,double *x,double *r)

	  

{

  mx_fill(r,2,2,1.,0.,0.,1.);

}



/***************
 * FUNCTION:     didtaz
 * PURPOSE:      jacobian of idta around z
 * ARGUMENTS:    double *z :
 *               double *x :
 *               double *r :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void didtaz(double *z,double *x,double *r)


{

  mx_fill(r,2,2,- 1.,0.,0.,- 1.);

}

/*************
  x[2] is an angle 
*************/



/***************
 * FUNCTION:     idta3
 * PURPOSE:      f = x - z in 2D where x[2] and z[2] 
                 are angles and need to be normalized
 * ARGUMENTS:    double *z :
 *               double *x :
 *               double *f :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void idta3(double *z,double *x,double *f)
{
  f[0] = x[0] - z[0];
  f[1] = x[1] - z[1];
  f[2] = norm_ang(x[2] - z[2],-M_PI);
  

}


/***************
 * FUNCTION:     didta3x
 * PURPOSE:      r = jacobian of idta3 around x
 * ARGUMENTS:    double *z :
 *               double *x :
 *               double *r :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


void didta3x(double *z,double *x,double *r)

{

  
  mx_fill(r,3,3,1.,0.,0.,
	        0.,1.,0.,
	        0.,0.,1.);

}


/***************
 * FUNCTION:     didta3z
 * PURPOSE:      r = jacobian of idta3 around z
 * ARGUMENTS:    double *z :
 *               double *x :
 *               double *r :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


void didta3z(double *z,double *x,double *r)

{

  mx_fill(r,3,3,-1.,0.,0.,
	        0.,-1.,0.,
                0.,0.,-1.);

}













