


/*
 *         Laser Data Segmentation
 *
 * two versions:
 *  segmentation(igl,option)
 * 
 *      input: an image of type "*igla"
 *          This image contain an array of igl->nbp
 *          raw points :  igl->tbp
 *      output: the same image with the list igl->lsgl of
 *           segment computed from the points igl->tbp.
 *
 *      option: 0 for least square estimation
 *           1 for kalman estimation
 *   
 *  get_sg_host(sgt,nbp,laser_val)
 *       
 *      input: the number of point "nbp" in the 
 *             laser pixel array "laser_val"
 *        
 *      output: the array of segment "sgt"
 *              and the number of segment (returned value)
 *      
 *
 *    Points are assumed coming from the same central position and
 *    ordered in angle in a trigonometric manner.
 *    
 *
 *   function conf_sg_host:
 *                         sets the parameters:
 *
 *                           thres (on the least square error)
 *                           nbmin (minimum number of points in
 *                                  a segment)
 *                           gap (maximum space between two consecutive
 *                                points in a segment)
 *                           max_range (maximum considered range)
 *
 *   function segmentation: computes the longest segments from
 *                          igl->tbp verifying the constraints set
 *                          by Set_segm_param
 *

 *   function first_discont_point       : returns TRUE if there is a discontinuity
 *                           between the  points j and k. If TRUE,
 *                           m is the discontinuous point.
 *   function first_continuous_point      : returns the index of the next
 *                              point, after k, which could be the
 *                              first point of a segment.
 *
 *   functions EndSS, eEndBP, EndBB :
 *                           computes the different kind of
 *                           intersections
 */


#include <stdio.h>
#include <math.h>
#ifdef VXWORKS
#define M_PI 3.141592653589793
#endif
#include <stdlib.h>

#include "definelocal.h"
#include "dbliste.h"
#include "lasdef.h"

/**** prototypes  ****/

#include "las_prot.h"
#include "lasegm_private_p.h"

#ifdef on_board

#include "laser.h"

#endif

#ifdef S500 /* Sensus 500 triangulation laser */
#define SATU 1.0 /* minimal radius of a point : filters points at the origin */

#define D_MIN 1.0 /* two points closer than that are considered close enough
		    to belong to the same segment */

#define D_MAX_INTER 3.0 /* maximum distance between to extremity points for intersection
			   computation */


/*********************
  Parameters for the segmentation
*********************/
static int step = 5; /* step for the split and merge */
static int nbmin = 4 ; /* minimum number of point in a segment (programable) */
static double thres = 0.2; /* threshold for the least square estimation (programable) */

static double  dist_points_max_square = 9.0; /* maximum distance between two points in a segment (inches)
		     used to detect discontinuities (programable) */

static double maxextend = 10.0 ; /* maximum extension of the segment beyond
			    the raw extremity points
			*/
static double max_range = 150.0;  /* maximum considered range of the laser system */

#endif /* S500 */

#ifndef S500

#define SATU 0.05 /* minimal radius of a point : filters points at the origin */

#define D_MIN 0.03 /* two points closer than that are considered close enough
		    to belong to the same segment */

#define D_MAX_INTER 0.1 /* maximum distance between to extremity points for intersection
			   computation */


/*********************
  Parameters for the segmentation
*********************/
static int step = 6; /* step for the split and merge */
static int nbmin = 6 ; /* minimum number of point in a segment (programable) */
static double thres = 0.009 ; /* threshold for the least square estimation (programable) */
static double max_range = 32.0;  /* maximum range of the laser system */

static double  dist_points_max_square = 0.01 ; /* maximum distance between two points in a segment (meters^2)
		     used to detect discontinuities (programable) */

static double maxextend = 0.0 ; /* maximum extension of the segment beyond
			    the raw extremity points
			*/

#endif

/* array of points to be treated */
/* (used by all the functions) */

static p2d *pointsloc;
static int nbpointsloc;




/*   
 *    nb (minimum number of points in
 *    a segment)
 *    max_gap = maximum distance between two consecutive points
 *              in meter in a same segment
 */


/***************
 * FUNCTION:     conf_sg_host
 * PURPOSE:      set parameters for the segmentation
 * ARGUMENTS:    double thr : threshold on the mean square error
                 Default = 0.009
		 If a segment has a higher mean error value
		 that thres, it will be rejected.
 *               int nb : minimum number of points in a segment
                 Default = 6         
 *               double max_gap : maximum distance between two points
                               in a segment( in meter).
		 Default = 0.1
 *               max_range : maximum range of the laser system

 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

int conf_sg_host(double thr,int nb,double max_gap, double range)

{
  if (range == 0) {
    printf ("SEGKIT lasegm.c : conf_sg_host null range !\n");
    return 0;
  }
  if (nb == 0) {
    printf ("SEGKIT lasegm.c : conf_sg_host null nb of points !\n");
    return 0;
  }
  thres = thr;
  nbmin = nb;
  dist_points_max_square = max_gap*max_gap;
  max_range = range;

  return 1;
}

/**********************

   ig = pointer on the image containing the array of points
	(ig->tbp).
   option for kalman filtering: 0 = least square only
				1 = grouping using
				    least square +
				    estimation using
				    Extended Kalman Filtering

   External call :  in Las_unc.c for Kalman estimation.

 ***********************/

/***************
 * FUNCTION:     segmentation
 * PURPOSE:      
 * ARGUMENTS:    igla *ig : ig = pointer on the image containing
                            the array of point (ig->tbp).
		 int op : option for kalman filtering: 
                                0 = least square only
				1 = grouping using
				    least square +
				    estimation using
				    Extended Kalman Filtering

 * ALGORITHM:    least square regression assuming points are
                 ordered in angle around 0,0 
		 If op = 1, at the very end, kalman estimation
		 using the grouping produced by the least square
		 algorithm.
		 returns always 0 (??)
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS:        las_segm_unc()/las_unc.c, first_continuous_point(),first_discont_point(),
                 alined(),stepping_back()  
 * CALLED BY:    
 ***************/


int segmentation (igla *ig,int op)


{
  int idt,first,last,attempt;
  int m,cut,disc;
  Sgl *psglc,*dte;
  struct matiner mic,mi0;
  int i;
  int steplocal;

  /* structure for the current segment */
  dte = Sgl_cree();

  pointsloc = ig->tbp ;
  nbpointsloc = ig->nbp ;

  /* usefull ? necessary ? */
  steplocal = step > nbmin ? nbmin : step;

  idt=0; /* index for the current segment */
  first = last = 0;

  first = first_continuous_point (last); /* first = index of the first point of dte
				   in the array of points
				*/
  /* initialization of the matrix of inertia (for least square) */
  mazmi(&mi0);
  mic = mi0;

  last = first;
  attempt = last+steplocal;
  if (attempt>nbpointsloc) attempt=nbpointsloc;
  psglc = NULL;

  while ( first < (nbpointsloc-1) )
    {
      cut = disc = FALSE;

      /* Looking for a discontinuity between last and attempt */

      if ( first_discont_point(last,attempt,&m) == TRUE)
	{
	  disc = TRUE;
	  attempt = m; /* discontinuity location; the point m can't
		    belong to the same segment as m-1*/
	}



      /* try to fit a line from first to attempt;
	 the segment first to last is already computed (if last != first);
	 The function tries to add last->attempt;
	 If it doesn't work, attempt steps back */


      if(disc == FALSE || attempt - first >= nbmin){
	while ( alined (first, last,attempt, &mic, dte)==FALSE)
	  {
	    attempt = stepping_back(first,last,attempt); /* steping back */
	    if (attempt == -1) {
	      for (i=0; i<nbpointsloc; i++) {
		printf ("%d %g %g %g\n", i, pointsloc[i].r, pointsloc[i].x,  pointsloc[i].y);
		return -1;
	      }
	    }
	    cut=TRUE;
	  }
      }
      
      /* a segment is finished */

      if ( cut || disc || (attempt==nbpointsloc) )
	{
#ifdef S500
	  if ( ((attempt-first) >= nbmin && pointsloc[attempt - 1].r > 40.)
	      || (attempt - first) >= nbmin + 3) /* if the number
					of point is large enough */
	    /* + special test when the last point is very close 
		 to the robot: we need more points so we
		 can discard the ones really close when we use
		 the kalman estimation
	      */
	    {
	      
#endif

#ifndef S500
	  if ( (attempt-first) >= nbmin ) /* if the number
					of point is large enough */
	    
	    {
	      
#endif

	     
	      /* we save dte in ig->lsgl */
	      psglc = Sgl_cree();
	      (*psglc) = (*dte);
	      psglc->ifirst= first ;
	      psglc->ilast= attempt-1;
	      psglc->o = atan2(psglc->vy,psglc->vx);
	      ig->lsgl = dlhcons(ig->lsgl ,psglc);
		

	      idt++;
	    }
	  mic = mi0;
	  if (cut == FALSE)
	    first = first_continuous_point(attempt);
	  else first = attempt-1;
	  last = first;
	  attempt = first+steplocal;
	  if (attempt>nbpointsloc) attempt=nbpointsloc;
	}
      else
	{
	  /* we keep trying to integrate points */
	  last = attempt;
	  attempt += steplocal;
	  if (attempt>nbpointsloc) attempt=nbpointsloc;
	}
    }



  Sgl_free (dte);

  ig->nsg = idt;


  /* uncertainty computation */

  if(op)
    las_segm_unc(ig);

  if (idt > 0) extremities(ig);

  
  
  return(0);
}



/***************
 * FUNCTION:     get_sg_host
 * PURPOSE:      segmentation using directly the server format
                 of laser values and returning a server format 
		 array of segments.

 * ARGUMENTS:    long *sgt : server format array of segment (output)
 *               long nbp : number of pixel points in laser_val (input)
 *               long *laser_val : server format array of pixels (input)

 * ALGORITHM:    1) transforms the "nbp" pixel points 
		 of laser_val into cartesian
		 data using the current 
		 selected calibration parameters
		 and fills the structure "ig"
		 2) calls segmentation()
		 3) copies the internal structure "ig" into
		 the array "sgt"
		 4) frees the internal structure.
		 if conf_sg_host  has been called like this:
		 conf_sg_host(a,nb,c,l), sgt must be an array 
		 of (4*nbp)/(nb-1) + 2 integers
		 
		 The function returns the number of segments.
		 It uses the function "segmentation" above.
		 
		 The number of segments will also be in sgt[0].
		 
		 Note: this segmentation will use only least square.

 * RETURN:       long : number of segments in "sgt"
 * SIDE EFFECT:  
 * CALLS: Imgl_cree()/las_struct.c, Imgl_free()/las_struc.c,
          Imgl_cart_fill()/las_struct.c, segmentation()
 * CALLED BY:    
 ***************/


static long get_sg_host(long *sgt,long nbp, long *laser_val )
{

  igla *ig;

  Sgl *sg;
  PDBL *l;
  long i;


  /*
     creating and filling the internal structure with
     current data
  */

  ig = Imgl_cree();
  Imgl_cart_fill(ig,nbp,laser_val);

  /* segmentation */

  segmentation(ig,0);

  /* number of segments */

  sgt[0] = ig->nsg;

  /* output interface */
  i = 1;

  for(l = ig->lsgl; l != NULL; l = dls(l)){
    sg = (Sgl *)dlel(l);
    sgt[i] = 10 * sg->first.x;
    sgt[i+1] = 10 * sg->first.y;
    sgt[i+2] = 10 * sg->last.x;
    sgt[i+3] = 10 * sg->last.y;
    i += 4;
  }

  /* destruction of ig */

  Imgl_free(ig);

  /* return */

  return(sgt[0]);

}




/***************
 * FUNCTION:     alined
 * PURPOSE:      determine weither the points from first to
                 last are alined (mean square error < thres).
		 Points first to j are already integrated in
		 the inertia matrix *mic.
		 The line parameters are computed in *dt (see lasdef.h).
		 Returns 1 if alined, 0 if not.
		 In case points are alined, updates the inertia matrix
		 "*mic".
		 NOTE: the inertia matrix doesn't have anything to do 
		 with the matrix representation in matrix.c 

 * ARGUMENTS:    int first : index of the first point integrated in *mic
 *               int last : index of the last point already integrated
 *               int attempt : index of the point
 *               struct matiner *mic : accumulation of first and second
                                       moments of inertia of points
				       belonging to the current segment
 *               Sgl *dt : current infinite line
 * ALGORITHM:    
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS: lsqr_approx()        
 * CALLED BY: segmentation()   
 ***************/


static int alined (int first,int last,int attempt,struct matiner *mic,
	Sgl *dt)

{
  struct matiner mi;

  int n,p;
  double err;



  mi = *mic;
  n = last-first;
  p = attempt-last;

  /* least square aproximation */

  lsqr_approx (&n,&mi,p,&pointsloc[last],dt,&err);



  if ( err>thres )
    {

      return(FALSE);
    }
  else
    {
      *mic = mi;
      return(TRUE);
    }


}


/***************
 * FUNCTION:     stepping_back
 * PURPOSE:      looks for the furthest point between pointsloc[last] and
                 pointsloc[attemp-1] from the line joining pointsloc[first]
		 to pointsloc[attempt].
		 This function is called only when the goup of points 
		 first->attempt has been determined as non-alined.
		 The furthest point returned  will then be taken as a new 
		 possible limit for the current segment.

 * ARGUMENTS:    int first :
 *               int last :
 *               int attempt :
 * ALGORITHM:    
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS:  dt2pt/lasutil.c      
 * CALLED BY: segmentation()   
 ***************/


static int stepping_back(int first,int last,int attempt)
{
  Sgl crd;
  int i,j,cut=attempt;
  double dm;
  double d;
  int max;
  
  if (pointsloc[first].x == pointsloc[attempt-1].x) {
    printf("sick bug while stepping_back: first = %d attempt = %d last = %d\n",
	   first, attempt, last);
    max = attempt>last?attempt:last;
    for (i=first; i<max; i++)
      printf ("%d  %5.4f %5.4f\n", i, pointsloc[i].x, pointsloc[i].y);
    return attempt+1;
  }

  dt2pt (&pointsloc[first], &pointsloc[attempt-1], &crd);
  
  /* crd = line passing through first and attempt-1 */
  j = last;
  dm = 0;
  if ( last == first )
    
    /* first can't be the farthest point from the cord */
    j++;
  else
    /* j-1 is the last point of the already computed segment;
       it is possible j-1 is the farthest point from the cord;
       */
    j--;
  
  /* looking for the furthest point from the cord */

  for (i=j;i<attempt-1;i++)
  {
    d = fabs (crd.vx*pointsloc[i].x + crd.vy*pointsloc[i].y + crd.d);
    if ( d > dm) { dm=d; cut =i+1; }
  }
  return(cut);
}





/***************
 * FUNCTION:     first_discont_point
 * PURPOSE:      looks if there is a discontinuity
                 between pointsloc[j] and pointsloc[k]
		 If yes, put the discontinuous point 
		 index in *m.
		 (means pointsloc[*m] cannot belong 
		 to the same segment as any points 
		 pointsloc[n], j <= n < *m)
		 Returns 1 if a discontunuity exists,
		         0 if not
		 A discontinuity between points i and n
		 occurs if:

		 - there is a local specularity angle of 
		   less than 0.5 degrees and the local difference in 
		   distance between i and n is more than D_MIN
		 or if:

		 - the distance between i and n is more than the gap
		   defined bye the user in conf_sg

		 or if:
		 
		 - the point n is at a distance less than SATU or
		   more than max_range, meaning this point can't be 
		   valid
		
 * ARGUMENTS:    int j :
 *               int k :
 *               int *m :
 * ALGORITHM:    
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY: segmentation()   
 ***************/

static int first_discont_point (int j,int k,int *m)

{
  int i;
  double thr,dtheta;
  double dist_square;

  for (i=j;(i<(nbpointsloc -1) && i<k );i++)
    {
      /* computing a theshold for specularity cheking */

      /* by testing if the angle (0-pointsloc[i],pointsloc[i]-pointsloc[i+1]) 
	 is more than 0.5 degrees (0.017 = tg(0.5)) */
      /* this is an aproximation */
      /* However, if the difference in distance is less than D_MIN, the test is         ok anyway */


      dtheta = fabs(norm_ang(pointsloc[i+1].o - pointsloc[i].o,-M_PI));

      /* specularity angle must be less than 0.5 degres */
      thr = dtheta*(pointsloc[i+1].r  + pointsloc[i].r) /0.017;

      /* XXXXXXX should be MIN here */
      thr = MAX(thr, D_MIN);

      dist_square = (pointsloc[i].x - pointsloc[i+1].x) *
		    (pointsloc[i].x - pointsloc[i+1].x) +
		    (pointsloc[i].y - pointsloc[i+1].y) *
		    (pointsloc[i].y - pointsloc[i+1].y) ;

      /* discontinuity : i+1 don't belong to i's segment */
      if ( dist_square  > dist_points_max_square 
	   || fabs(pointsloc[i].r - pointsloc[i+1].r) > thr  ) {
	  *m = i+1; 
	  return(1);
 	}
      /* pointsloc[i] is too near or too far : 
	 don't belong to any segment */
      if ( pointsloc[i].r < SATU  || pointsloc[i].r >= max_range) { 
	*m = i;
	return(1);
      }
    }

  return(0);
}



/***************
 * FUNCTION:     first_continuous_point
 * PURPOSE:      returns the index of the next
                 point, after k, which could be the
		 first point of a segment.
		 This function looks for a point such that 
		 the following point is continuous to it, in the same
		 sens than first_discont_point right above.
 * ARGUMENTS:    int k :
 * ALGORITHM:    
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:  segmentation()  
 ***************/

static int first_continuous_point(int k)

{
  int i;
  double dist_square,thr,dtheta ;

  for (i=k;i<(nbpointsloc-1);i++)
    {
      dtheta = fabs(norm_ang(pointsloc[i+1].o - pointsloc[i].o,-M_PI));

      /* specularity angle must be less than 0.5 degres */
      thr = dtheta*(pointsloc[i+1].r  + pointsloc[i].r) /0.017;

      /* XXXXXXX should be MIN here, no ? */
      thr = MAX(thr, D_MIN);

      /* square distance between the 2 points */
      dist_square = (pointsloc[i].x - pointsloc[i+1].x) * 
		    (pointsloc[i].x - pointsloc[i+1].x) +
		    (pointsloc[i].y - pointsloc[i+1].y) *
		    (pointsloc[i].y - pointsloc[i+1].y);

      if (/* cartesian gap small enough */
	  dist_square <= dist_points_max_square 
	  /* radius gap _or_ specularity small enough */
	  && fabs(pointsloc[i].r - pointsloc[i+1].r) <= thr 
	  /* point not too close to the sensor */ 
	  && pointsloc[i].r >= SATU 
	  /* point not too far from the sensor */
	  && (pointsloc[i].r < max_range))
	return(i);
    }
  return(nbpointsloc);
   
}


/***************
 * FUNCTION:     extremities
 * PURPOSE:      least mean square estimation and kalman
                 compute only the supporting line of the segments
		 This function computes the extremities.
              
 * ARGUMENTS:    igla *ig :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:   EndBlast(),EndBfirst(),EndVV_BB()
 * CALLED BY: segmentation()        
 ***************/

static void extremities(igla *ig)

{
  PDBL *list; /* list of infinite lines */

  Sgl *psglc, /* pointer on the current segment */
      *psglp; /* " "            precedent segment */




  list = ig->lsgl;
  pointsloc = ig->tbp;


  psglc = NULL;
  psglp = (Sgl *)(dlel(list));
  list = dls(list);

  /* last extremity of the last segment
    (the segments are in inverse order in the liste)
  */

  if (psglp->ilast == (ig->nbp -1))
    {
      psglp->typl = 'O'; /* O means "open" : there is no information before
			    to tell us the semantic of this point */

      EndBlast (psglp);
      
    }
  else
    {
      psglp->typl = 'B'; /* B means "border" : it is the begining of the 
			    visible part of the segment, but it is not
			    a vertex
			*/
      EndBlast(psglp);
    }

  for(;list != NULL;list = dls(list))
    {
      psglc = psglp;
      psglp = (Sgl *)(dlel(list));

      /*
	 if the first point of psglc is "close" ( in term
	 of angle) to the last point
	 of pslgp: test if there is a vertex.
      */
      if ((psglc->ifirst - psglp->ilast) <= 1)
	EndVV_BB (psglp, psglc);
      /* otherwise, psglc->first and psglp->last are borders */
      else
	EndBB (psglp, psglc);
    }

  /* first of the first segment */
  /* same remarks as for the last of the last segment */
  if (psglp->ifirst == 0)
    {
      psglp->typf = 'O';
      EndBfirst (psglp);
    }
  else
    {
      psglp->typf = 'B';
      EndBfirst (psglp);
    }

 
  
}

/*****
  uses function of lasutil.c for intersecting segments
*****/


/***************
 * FUNCTION:     EndVV_BB
 * PURPOSE:      tests if the intersection between psglp
                 and psglc is "reasonably" a vertex. 
		 By reasonable, we mean:
		 - the end of psglp and the begining of psglc
		   are close (angle < 3.5 degrees and 
		   distance < 3 inches)
		 - the intersection point is close to the mean 
		 angle between the two extremities (no more than 5.5 degrees)

		 If the conditions are fulfilled, 
		      psglp->last and psglc->first are set to be the same
		      psglp->slast points to psglc
		      psglc->sfirst points to psglp
		      psglp->typl = psglc->typf = 'V'

 * ARGUMENTS:    Sgl *psglp :
 *               Sgl *psglc :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        EndBB(), intersg()
 * CALLED BY:    extremities()
 ***************/


static void EndVV_BB (Sgl *psglp, Sgl *psglc)

{
  double o,diff,allo,alli;
  p2d pt;

  o = (pointsloc[psglp->ilast].o + pointsloc[psglc->ifirst].o)/2;
  o = norm_ang(o,-M_PI);

  /* o = mean angle between the two extremities */ 

  intersg(psglp,psglc,&pt);
  
  /* pt = intersection point */

  alli = fabs(norm_ang(atan2(pt.y,pt.x) - o,-M_PI));

  /* alli = difference in angle between the intersection point
     and the mean angle o */

  diff = fabs(norm_ang(pointsloc[psglp->ilast -1].o - pointsloc[psglc->ifirst + 1].o,-M_PI))/2;
  
  /* diff = half angle distance between the two extremities */

  allo = fabs((pointsloc[psglp->ilast].r - pointsloc[psglc->ifirst].r));
  
  /* allo = difference in distance between the two extremities */

  /* test if the intersection is "reasonable" */
     if ( diff > 0.03 || (alli > diff*4 && alli > 0.1) || allo > D_MAX_INTER ) 
    {
      /* not reasonable */
      EndBB (psglp, psglc);
    }

  /* V-V */
  /* ok, pt is a vertex, we link the two segments together */
  else
    {
      psglp->typl = psglc->typf = 'V';
      psglp->slast = psglc;
      psglc->sfirst = psglp;
      (psglp->last).x = (psglc->first).x = pt.x;
      (psglp->last).y = (psglc->first).y = pt.y;
      (psglp->last).r = (psglc->first).r = sqrt(pt.x * pt.x + pt.y * pt.y);
      

    }
}


/***************
 * FUNCTION:     EndBB
 * PURPOSE:      computes the border points of
                 the end of psglp and the begining 
		 of psglc
 * ARGUMENTS:    Sgl *psglp :
 *               Sgl *psglc :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        EndBlast(), EndBfirst()
 * CALLED BY:    EndVV_BB()
 ***************/

static void EndBB (Sgl *psglp,Sgl *psglc)
 
 
{
  psglp->typl = psglc->typf = 'B';
  EndBfirst (psglc);
  EndBlast (psglp);
}



/***************
 * FUNCTION:     EndBfirst
 * PURPOSE:      computes the first Border extremity of the segment 
                 sg 
 * ARGUMENTS:    Sgl *sg :
 * ALGORITHM:    There are two ways to compute the extremity
                 of the segment. They are obviously both related
		 to the first point of sg in the array pointsloc[].
		 Let us call it pd (as in ths code below).
		 
		 Since sg has been estimated by least square or
		 kalman, pd does not exactly belong to sg.
		 Theoretically, the begining of sg should be 
		 the intersection of the line joining the center of 
		 the robot and the infinite line defined in sg.
		 However, sometimes, due to specular angle and inaccuracy
		 on the point and the line, this intersection can go very 
		 far from pd creating very long fake lines. 
		 The static variable maxextend sets the acceptable limit
		 (here = sqrt(15) inches).
		 If the intersection go further, then we simply take
		 the orthogonal projection of pd on sg as the first
		 extremity of sg.
		 
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

static void EndBfirst (Sgl *sg)
     
{
  Sgl dtt;
  p2d *pi, *pd;
  double o,l;

  pi = &(sg->first);
  pd = &(pointsloc[sg->ifirst]);
  o = pointsloc[sg->ifirst].o ;

  /** dtt = line from robot center to pd **/
  dtt.vx = -sin(o); dtt.vy = cos(o); dtt.d = 0;
  
  /** if maxextend non null **/
  if (maxextend > 0.000000001) {
    /* sg->first = intersection sg/dtt **/
    intersg(sg,&dtt,pi);
    /* l = square of distance pd<->sg->first */  
    l = (pi->x- pd->x)*(pi->x - pd->x)+ (pi->y-pd->y)*(pi->y-pd->y);
    if (l > maxextend) 
   
      /* if l two big -> orthogonal projection */
      proj_orth (sg, pd, pi);
  } else
      proj_orth (sg, pd, pi);

  pi->o = atan2(pi->y,pi->x);
  pi->r  = sqrt(pi->x * pi->x +
		      pi->y * pi->y);
}


/***************
 * FUNCTION:     EndBlast
 * PURPOSE:      computes the last Border extremity of
                 the segment sg.
 * ARGUMENTS:    Sgl *sg :
 * ALGORITHM:    see EndBfirst (replace first by last)
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

static void EndBlast (Sgl *sg)

{
  Sgl dtt;
  p2d *pi, *pd;
  double o,l;

  pi = &(sg->last);
  pd = &(pointsloc[sg->ilast]);
  o = pointsloc[sg->ilast].o;
  dtt.vx = -sin(o); dtt.vy = cos(o); dtt.d = 0;

  if (maxextend > 0.0000001) {
      
    intersg(sg,&dtt,&(sg->last));
      l = (pi->x- pd->x)*(pi->x - pd->x)+ (pi->y-pd->y)*(pi->y-pd->y);
      if (l > maxextend)
	  proj_orth (sg, pd, pi);
  } else
      proj_orth (sg, pd, pi);

  (sg->last).o = atan2(sg->last.y,sg->last.x);
  sg->last.r  = sqrt(sg->last.x * sg->last.x +
		      sg->last.y * sg->last.y);
}
   

/***************
 * FUNCTION:     mazmi
 * PURPOSE:      zeroes the inertia matrix mi
 * ARGUMENTS:    struct matiner *mi :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

static void mazmi(struct matiner *mi)

{
  mi->sx = mi->sy = mi->sx2 = mi->sy2 = mi->sxy = 0;
}




/**************
  computes a least-square estimation of the points
  from tab. 
  n is the number of points already
  integrated;

  mi is the matrix of inertia of those points

  p is the number of points to integrate

  tab is the array of points to be integrated

  dte is the line we want to be computed

  err is the quadratic error;

  Side effects:

  dte (->vx,->vy and ->d, ->er) is updated such that d > 0
  and  || vx^2 + vy^2 || = 1;

  mi is updated;

***************/
  

/***************
 * FUNCTION:     lsqr_approx
 * PURPOSE:      computes a least square line estimation of the points
                 from tab[0] to tab[p-1] included.
		 the line equation is supposed to be
		 vx*x + vy*y + d = 0
		 with 
		 |vx^2 + vy^2| = 1
		 This function computes vx,vy and d.
		 dte->vx, dte->vy and dte->d are filled accordingly.
		 dte->er is filled with the mean square error
 * ARGUMENTS:    int *n :
 *               struct matiner *mi :
 *               int p :
 *               p2d *tab :
 *               Sgl *dte :
 *               double *err :
 * ALGORITHM:    The algorithm is based on lagrange minimization
                 of a criteria under constraints.
		 We define the error of each point (x_i,y_i) with respect
		 to the final estimation as :
		      e_i = vx * x_i + vy * y_i + d
		 the least square criteria is then 
		      J = sum_0^n(e_i)^2
		 Minimizing this criteria over the 3d space (vx,vy,d) under
		 the constraint
		      |vx^2 + vy^2| = 1
		 is equivalent to minimize
		      J2 = J + lamda*(1 - vx^2 + vy^2)
		 under the 4d space (vx,vy,d,lamda) with no constraint
		 
		 It is easy then, to see that lambda is an eigenvalue
		 of the matrix:
		 
		 | sum(x_i^2)-sum(x_i)^2     sum(x_i*y_i)|
		 | sum(x_i*y_i)     sum(y_i^2)-sum(y_i)^2|
		 
		 It is also easy to see that the actual value of
		 the ctriteria is equal to lambda.

		 vx, vy is then the normalized eigenvector

		 The main point of all this is that it is
		 enough to keep memory of 
		 sum(x_i), sum(y_i), sum(x_i^2), sum(y_i^2)
		 and sum(x_i*y_i).

		 That is what the structure "matiner" is made
		 for.
		 From this, we can describe the process:

		 mi contains memory of points already integrated in 
		 the current segment being estimated.
		 n in the number of points already integrated
		 p is the number of points to be integrated
		 tab is the array of points to be integrated
		 dte is the resulting line (output)
		 err is the resulting mean square error.
		 mi will be updated.
		 
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

static void lsqr_approx(int *n,struct matiner *mi,int p,p2d *tab,Sgl *dte,double *err)

{
      double vx,vy,d,rdelta,lam,v;
      double gx,gy,gx2,gy2,gxy;
      double x,y,q;
      int i;

/* loop over  the points */
      /* adding the points in mi */
      if (p>0)
	{
	  for(i=0;i<p;i++)
	    {
	      x = tab[i].x;
	      y = tab[i].y;
	      
	      mi->sx += x;
	      mi->sy += y;
	      mi->sx2 += x*x;
	      mi->sy2 += y*y;
	      mi->sxy += x*y;
	    }
	  *n += p;
	}
      /* we could also remove those points */
      else
	{
	  p = -p;
	  for(i=0;i<p;i++)
	    {
	      x = tab[i].x;
	      y = tab[i].y;
	      
	      mi->sx -= x;
	      mi->sy -= y;
	      mi->sx2 -= x*x;
	      mi->sy2 -= y*y;
	      mi->sxy -= x*y;
	    }
	  *n -= p;
	}
      

      /* computing the matrix coeeficients */

         gx = mi->sx/(*n);
         gy = mi->sy/(*n); 
         gx2 = mi->sx2/(*n);
         gy2 = mi->sy2/(*n);
         gxy = mi->sxy/(*n);

         gx2 = gx2-gx*gx;
         gy2 = gy2-gy*gy;
         gxy = gxy-gy*gx; 



      rdelta = sqrt ( (gx2-gy2)*(gx2 - gy2) + 4*(gxy*gxy) );
      /* 
	 warning, here lam does not represent the smallest 
	 eigen value (see "q" below). It is a trick
	 for numerical sake
      */
         lam = (gx2+gy2+rdelta)/2.;
      /* computing the eigen vector */
         if  ((fabs(lam-gx2))<=(fabs (lam-gy2))){
            vy = lam-gy2;
            vx = -gxy;
          } else {
            vy = gxy;
            vx = -lam+gx2;
         } 

      /* d comes out easily from the line equation */
      d = -vx*gx - vy*gy;

      /* normalizing and setting d to be positive */
      v = sqrt(vx*vx + vy*vy);
      if (d<0) v= -v;
      vx /= v;
      vy /= v;
      d /= v;

      /* the mean square error = smallest eigen value */
      q = (gx2+gy2-rdelta)/2;
      if (q<0) 
	q=0;
      q = sqrt(q);
      *err = q;
      /* filling dte fields */
      dte->vx = vx;
      dte->vy = vy;
      dte->d = d;
      dte->er = q;
    }




























