/****************************
Little (but useful !) functions 
dealing with points and segments

NOTE: p2d and Sgl structures are described in lasdef.h
****************************/

#include <stdio.h>
#include <math.h>
#ifdef VXWORKS
#define M_PI 3.141592653589793
#endif
#include <stdlib.h>

#include "definelocal.h"
#include "dbliste.h"
#include "lasdef.h"
#include "las_struct_p.h"
#include "lasutil_p.h"



/***************
 * FUNCTION:     norm_ang
 * PURPOSE:      normalizes an angle between p and 2*Pi + p
 * ARGUMENTS:    double a :
 *               double p :
 * ALGORITHM:    
 * RETURN:       double : an angle between p and 2*M_PI + p
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

double norm_ang(double a,double p)
   
{
  double b;

/*   b = fmod(a,2*M_PI); */
  b = a;
  while(b <= p)
    b += 2*M_PI;

  while(b > p + 2*M_PI)
    b -= 2*M_PI;

  return(b);
}


/***************
 * FUNCTION:     intersg
 * PURPOSE:      computes the intersection between 
                 s1 and s2. Puts the result in p.
                 Returns 0 if segments are parallel
                 Otherwise, returns 1
 * ARGUMENTS:    Sgl *s1 :
 *               Sgl *s2 :
 *               p2d *pt :
 * ALGORITHM:    
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

int intersg(Sgl *s1,Sgl *s2,p2d *pt)
    
{
  double det;

  if (s1 == s2) {
    printf("segments are identical !\n");
    return(0);
  }
    
  det = s1->vx*s2->vy - s1->vy*s2->vx;
  
  if(fabs(det) < 1.0e-4) {
/*     printf("segments are parallel !\n"); */
    return(0);
  }
  pt->x = (s2->d * s1->vy - s1->d * s2->vy)/det;
  pt->y = (s2->vx * s1->d - s1->vx * s2->d)/det;
  return(1);
}



/***************
 * FUNCTION:     proj_orth
 * PURPOSE:      computes the orthogonal projection 
                 of pt on dt.
                 Result in ptp (which has to be 
		 allocated before the call)
 * ARGUMENTS:    Sgl *dt :
 *               p2d *pt :
 *               p2d *ptp :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


void proj_orth(Sgl *dt,p2d *pt,p2d *ptp)

{
Sgl dtt;

dtt.vx = dt->vy;
dtt.vy = -dt->vx;
dtt.d = -dtt.vx * pt->x - dtt.vy * pt->y;

intersg(dt,&dtt,ptp);

}

/***************
 * FUNCTION:     dt2pt
 * PURPOSE:      Initialize a line dt
                 from 2 points pt1,pt2
                 (computes dt->vx,dt->vy,dt->d)
                 vx,vy is the normal with the convention that 
		 it goes to the left when one walk along from pt1 to pt2.
		 d = -x*vx -y*vy where (x,y) is any point belonging
		 to dt. 
		 In particular:
		 d = -pt1->x*dt->vx - pt1->y*dt->vy

 * ARGUMENTS:    p2d *pt1 :
 *               p2d *pt2 :
 *               Sgl *dt :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void dt2pt (p2d *pt1,p2d *pt2,Sgl *dt)
   
{
  double no;
  
  dt->vx = pt1->y - pt2->y;
  dt->vy = pt2->x - pt1->x;
  no =  sqrt( dt->vx*dt->vx + dt->vy*dt->vy);
  dt->vx /= no; dt->vy /= no;
  dt->d = -dt->vx*pt1->x - dt->vy*pt1->y;
}




/***************
 * FUNCTION:     sg_calc_param
 * PURPOSE:      computes vx,vy,d from extremity points
                 such that ||vx^2 + vy^2|| = 1 and the vector 
                 vx,vy goes on the left side of sg when going from
                 the begining to the end;
		 Basically the same as dt2pt but assuming 
		 pt1 and pt2 are already in sg->first and sg->last
		 respectively.
		 Computes also the polar representation ->r and ->o
		 of sg.
 * ARGUMENTS:    Sgl *sg :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void sg_calc_param(Sgl *sg)
     

{

  double x1,y1,x2,y2,norm;
  
  x1 = (sg->first).x;
  y1 = (sg->first).y;

  x2 = (sg->last).x;
  y2 = (sg->last).y;

  norm = sqrt((x1 - x2)*(x1-x2) + (y1-y2)*(y1-y2));

  sg->vx = (y1 - y2)/norm;
  sg->vy = (x2 - x1)/norm;

  

  sg->d = sg->r = -(sg->vx * x1 + sg->vy * y1);

 
  sg->o = atan2(sg->vy, sg->vx);

}


/***************
 * FUNCTION:     InsidePoly
 * PURPOSE:      tests if the point defined by
                 x = pos[0], y = pos[1] is inside
		 the polygone define by the segment list
		 poly->lsgl.
		 Returns 1 if inside
		         0 if outside
                         -1 if poly = NULL or
			    if the list of segments is not
			    closed.
			 
 * ARGUMENTS:    double *pos :
 *               igla *poly :
 * ALGORITHM:    
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

int InsidePoly(double *pos,igla *poly)
{
  double ag1,ag2,agtot = 0;
  Sgl *s,*sb;
  double x1,y1,x2,y2;


  if(poly == NULL)
    return(-1);
  if(poly->lsgl == NULL)
    return(-1);

  sb = (Sgl *)dlel(poly->lsgl);

  s = sb;

  x1 = s->first.x - pos[0];
  y1 = s->first.y - pos[1];

  ag1 = atan2(y1,x1);
  
  do{
    if(s->typl != 'V')
      {
	printf("InsidePoly function: polygon not closed !\n");
	return(-1);
      }
    s = s->slast;
    x2 = s->first.x - pos[0];
    y2 = s->first.y - pos[1];
  
    ag2 = atan2(y2,x2);
    
    

    agtot += norm_ang(ag2 - ag1,-M_PI);

    ag1 = ag2;
    x1 = x2;
    y1 = y2;

  }while(s != sb);

  if(fabs(agtot) < 0.1)
    return(0); /* outside */

  return(1); /* inside */

}
  
  











