/***********************************************************************
 *     
 *     Extended Kalman Filtering for implicit measurement equations
 *
 *     Observation equation fonc(y,x) = 0 
 *     y= Observation , x= State, 
 *       
 *     dimf, dimy, dimx : dimensions of func,y and x.
 *     vy = Variance on y( matrix dimy*dimy),
 *     vx = Variance on x (matrix dimx*dimx)
 *     func,dfuncy,dfuncx : 
 *                Observation fonction, Jacobian of func about y and 
 *                Jacobian of func about x
 *      
 *       xp,vxp : previous estimate of x and its variance
 *       
 *       xn,vxn : output = new estimate and its variance 
 *     
 *       param is a vector of double wich can be used as a non-uncertain vector of
 *       parameters in the Observation function and its jacobians.
 *       
 *       vxp, vxn and vy are symmetrical matrices as defined in matrix.c in ../util
 *          
 *       Remark: &xp and &vxp may be equal to &xn and &vxn respectively.
---------------------------------------------------------------------- */

/******* ALL REFERENCES TO DOCUMENTATION REFER TO THE FILE "kalman.ps" */


#include <stdio.h>
#include <stdlib.h>
/* #include <assert.h> */



#define MAX(x,y) ( x > y ? x : y )
#include "mat.h"
#include "kal_prot.h"



static double *temp=NULL;
static double *gamma,*K;
static long lastdimf,lastdimx,lastdimy;
static double *f,*jacy,*jacx;
static double *txpp,*txll,*xntp,*vxntp,*vxntpp;
static double *W,*W1,*W2,*W3;


/***************
 * FUNCTION:     fusion
 * PURPOSE:      computes a Kalman estimation of xp and vxp using 
                 the observation function func() and the observation vector
                 y and vy. 
		 Puts the results in xn and vxn.
		 You may do xn = xp and vxn = vxp.
 * ARGUMENTS:    void (*func) () : pointer to the Observation function
 *               void (*dfuncy) () : pointer to a function computing J_y
 *               void (*dfuncx) () : pointer to a function computing J_x
 *               double *y : Observation Vector 
 *               double *vy : Covariance matrix of y
 *               double *xp : input State Vector
 *               double *vxp : Covariance matrix of xp
 *               double *param :  additional parameters to func()(see doc) 
 *               double *xn : output State Vector
 *               double *vxn : Covariance matrix of xp
 *               long dimf : dimension of func()
 *               long dimy : dimension of y
 *               long dimx : dimension of xp and xn
 * ALGORITHM:    computes in the order (look at the documentation):
                 equation 8
		 jacobians J_y and J_x
		 equation 9
		 equation 10
		 equation 11
		 equation 12 
		 equation 13
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        matrix functions in matrix.c
 * CALLED BY:    
 ***************/


void fusion (void (*func)(),void (*dfuncy)(),void (*dfuncx)(),
	     double *y,double *vy,double *xp,double *vxp,
	     double *param,double *xn,double *vxn,
	     long dimf,long dimy,long dimx)

{


    /* allocation of temporary memory */

  Kal_tempo(dimf,dimy,dimx);

  /* Computation of the Jacobians and the Observation funtion */

  if(param == NULL){
    dfuncy(y,xp,jacy);
    dfuncx(y,xp,jacx);
    func(y,xp,f);
  }

  else {
    dfuncy(y,xp,param,jacy);
    dfuncx(y,xp,param,jacx);
    func(y,xp,param,f);
  }


  /* Variance on the Observation = gamma */

  mxfs_mmulttr (jacx,vxp,gamma,dimf,dimx,dimx,dimx);
  mxfs_mmulttr (jacy,vy,W,dimf,dimy,dimy,dimy);
  mxs_plus (gamma,W,gamma,dimf,dimf);
  mxs_inv (gamma,W,dimf,dimf);

  /* remark : here , W= inv(gamma) */

  /*  Kalman's Gain computation */

  mx_trans (jacx,W3,dimf,dimx);
  mxsf_mult(vxp,W3,W1,dimx,dimx,dimx,dimf);
  mxfs_mult(W1,W,K,dimx,dimf,dimf,dimf);

  /* New estimate .... */

  mx_mult (K,f,xntp,dimx,dimf,dimf,1);
  mx_diff (xp,xntp,xn,dimx,1);

  /* and its variance */

  mx_mult (K,jacx,W2,dimx,dimf,dimf,dimx);
  mxfs_mult(W2,vxp,vxntp,dimx,dimx,dimx,dimx);
  mxfs_copy (vxntp,vxntpp,dimx,dimx);
  mxs_diff (vxp,vxntpp,vxn,dimx,dimx);


}

/** 
  variables for memorizing input/output in Dist_maha() in case we want
  to do Dfusion() afterwards
**/

static double *pxp,*pvxp,*pxn,*pvxn ;
static long pdimf,pdimx;

/***************
 * FUNCTION:     Dist_maha
 * PURPOSE:      Computes the likelyhood of the relationship "func()"
                 between xn and y (see documentation).
 * ARGUMENTS:    void (*func) () : pointer to the Observation function
 *               void (*dfuncy) () : pointer to a function computing J_y
 *               void (*dfuncx) () : pointer to a function computing J_x
 *               double *y : Observation Vector 
 *               double *vy : Covariance matrix of y
 *               double *xp : input State Vector
 *               double *vxp : Covariance matrix of xp
 *               double *param :  additional parameters to func()(see doc) 
 *               double *xn : output State Vector
 *               double *vxn : Covariance matrix of xp
 *               long dimf : dimension of func()
 *               long dimy : dimension of y
 *               long dimx : dimension of xp and xn
 * ALGORITHM:    computes (see documentation)
                 equation 8
                 jacobians J_y and Y_x
		 equation 9
		 returns result of equation 14
 * RETURN:       double: equation 14
 * SIDE EFFECT:  
 * CALLS:        matrix functions in matrix.c
 * CALLED BY:    
 ***************/


double Dist_maha (void (*func)(),void (*dfuncy)(),void (*dfuncx)(),
	          double *y,double *vy,double *xp,double *vxp,
	          double *param,double *xn,double *vxn,
	          long dimf,long dimy,long dimx)

{
  double d2;

    /* allocation of temporary memory */

  Kal_tempo(dimf,dimy,dimx);

    /* memorization so you could call Dfusion right after
       this function */

    pxp = xp; pvxp = vxp; pxn = xn; pvxn = vxn;
    pdimx = dimx; pdimf = dimf;
  /* calcul des jacobiennes (linearisation autour de y et xp) */
  if(param == NULL){
    (*dfuncy)(y,xp,jacy);
    (*dfuncx)(y,xp,jacx);
    (*func)(y,xp,f);
  }

  else {
    dfuncy(y,xp,param,jacy);
    dfuncx(y,xp,param,jacx);
    func(y,xp,param,f);
  }


  /* calcul de gamma */

  mxfs_mmulttr (jacx,vxp,gamma,dimf,dimx,dimx,dimx);
  mxfs_mmulttr (jacy,vy,W,dimf,dimy,dimy,dimy);
  mxs_plus (gamma,W,gamma,dimf,dimf);
  mxs_inv (gamma,W,dimf,dimf);

  /* calcul de la distance de Mahalanobis */

  mxfs_trmultm (f,W,&d2,dimf,1,dimf,dimf);
  return(d2);



}


/***************
 * FUNCTION:     Dfusion
 * PURPOSE:      computes a Kalman estimate of the 
                 State vector in pxp with variance vpxp.
		 Uses the prediction error in the variable "f"
		 and the jacobians in jacx and jacy.
		 Puts the result in xntp and vxntp.
		 Warning: this function assumes the static variables
		 f,jacx,jacy,gamma and W already computed by Dist_maha()
		 useful to proceed to a Dist_maha test before without recomputing
		 this stuff.
		 xntp and vxntp must contain the pointers to the vector/matrix you 
		 want to receive the result. Normally, Dist_maha() sets:
		 xntp = xn and vxntp = vxn.
		 Same thing for the input, Dist_maha() sets:
		 pxp = xp and vpxp = vxp
 * ARGUMENTS:    None
 * ALGORITHM:    computes(see documentation)
                 equation 10
		 equation 11
		 equation 12
		 equation 13
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        matrix funxtions in matrix.c
 * CALLED BY:    
 ***************/

void Dfusion (void)

{


  /* gamma = variance on the Observation */

  /* W= inv(gamma) */

  /*Kalman's Gain  = K */

  mx_trans (jacx,W3,pdimf,pdimx);
  mxsf_mult(pvxp,W3,W1,pdimx,pdimx,pdimx,pdimf);
  mxfs_mult(W1,W,K,pdimx,pdimf,pdimf,pdimf);

  /* New estimate ... */
  mx_mult (K,f,xntp,pdimx,pdimf,pdimf,1);
  mx_diff (pxp,xntp,pxn,pdimx,1);

  /* and its variance */

  mx_mult (K,jacx,W2,pdimx,pdimf,pdimf,pdimx);
  mxfs_mult(W2,pvxp,vxntp,pdimx,pdimx,pdimx,pdimx);
  mxfs_copy (vxntp,vxntpp,pdimx,pdimx);
  mxs_diff (pvxp,vxntpp,pvxn,pdimx,pdimx);


}




/***************
 * FUNCTION:     Test_fus
 * PURPOSE:      Tests if the "likelyhood" value of the Observation is less
                 than a threshold. If yes, proceeds to the filtering.
 * ARGUMENTS:    void (*func) () : pointer to the Observation function
 *               void (*dfuncy) () : pointer to a function computing J_y
 *               void (*dfuncx) () : pointer to a function computing J_x
 *               double *y : Observation Vector 
 *               double *vy : Covariance matrix of y
 *               double *xp : input State Vector
 *               double *vxp : Covariance matrix of xp
 *               double *param :  additional parameters to func()(see doc) 
 *               double *xn : output State Vector
 *               double *vxn : Covariance matrix of xp
 *               long dimf : dimension of func()
 *               long dimy : dimension of y
 *               long dimx : dimension of xp and xn
 *               double thres :likelyhood threshold 
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        Dist_maha() and Dfusion()
 * CALLED BY:    
 ***************/

long Test_fus (void (*func)(),void (*dfuncy)(),void (*dfuncx)(),
	       double *y,double *vy,double *xp,double *vxp,
	       double *param,double *xn,double *vxn,
	       long dimf,long dimy,long dimx, double thres)

{
  double d2;

  d2 = Dist_maha(func,dfuncy,dfuncx,y,vy,xp,vxp,param,xn,vxn,dimf,dimy,dimx);

  if(d2 < thres){
    Dfusion();
    return(1);
    }
  return(0);

}


/********
little basic functions for identity.
One can use them as parameters each time the
Observation function is an identity ( y = x)
Written till dimx = dimy = dimf = 3.
********/


/***************
 * FUNCTION:     idt1
 * PURPOSE:      puts x - y in r
 * ARGUMENTS:    double *y :
 *               double *x :
 *               double *r :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void idt1(double *y,double *x,double *r)
    
{

  *r = *x - *y;

}


/***************
 * FUNCTION:     didtx1
 * PURPOSE:      r = jacobian of idt1 with respect to x
 * ARGUMENTS:    double *y :
 *               double *x :
 *               double *r :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void  didtx1(double *y,double *x,double *r)

{

  *r = 1.;

}


/***************
 * FUNCTION:     didty1
 * PURPOSE:      r = jacobian of idt1 with respect to y
 * ARGUMENTS:    double *y :
 *               double *x :
 *               double *r :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


void didty1(double *y,double *x,double *r)
{

  *r = -1.;

}

/***************
 * FUNCTION:     idt2
 * PURPOSE:      puts x - y in r (dim = 2)
 * ARGUMENTS:    double *y :
 *               double *x :
 *               double *r :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void idt2(double *y,double *x,double *r)
{

  mx_diff(x,y,r,2,1);

}

/***************
 * FUNCTION:     didtx2
 * PURPOSE:      (dim = 2) r = jacobian of idt2 with respect to x
 * ARGUMENTS:    double *y :
 *               double *x :
 *               double *r :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void didtx2(double *y,double *x,double *r)

{

  mx_fill(r,2,2,1.,0.,0.,1.);

}

/***************
 * FUNCTION:     didty2
 * PURPOSE:      (dim = 2) r = jacobian of idt2 with respect to y
 * ARGUMENTS:    double *y :
 *               double *x :
 *               double *r :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


void didty2(double *y,double *x,double *r)
{

  mx_fill(r,2,2,- 1.,0.,0.,- 1.);

}

/***************
 * FUNCTION:     idt3
 * PURPOSE:      puts x - y in r (dim = 3)
 * ARGUMENTS:    double *y :
 *               double *x :
 *               double *r :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void idt3(double *y,double *x,double *r)
{

  mx_diff(x,y,r,3,1);

}

/***************
 * FUNCTION:     didtx3
 * PURPOSE:      (dim = 3) r = jacobian of idt3 with respect to x
 * ARGUMENTS:    double *y :
 *               double *x :
 *               double *r :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void didtx3(double *y,double *x,double *r)

{

  mx_fill(r,3,3,1.,0.,0.,
	  0.,1.,0.,
	  0.,0.,1.);

}

/***************
 * FUNCTION:     didty3
 * PURPOSE:      (dim = 3) r = jacobian of idt3 with respect to y
 * ARGUMENTS:    double *y :
 *               double *x :
 *               double *r :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void didty3(double *y,double *x,double *r)

{

  mx_fill(r,3,3,- 1.,0.,0.,
	  0.,- 1.,0.,
	  0.,0.,-1.);

}


/***************
 * FUNCTION:     Kal_tempo
 * PURPOSE:      All local matrices are locally static so we
                 can optimize memory allocation. Memory is reallocated
		 only if one of the dimensions dimf, dimx or dimy gets
		 bigger than the biggest one so far.
		 This implementation allows for the Dfusion() function
		 by keeping in memory intermediate computations.
 * ARGUMENTS:    long dimf : dim of the Observation Function
 *               long dimy : dim of the Observation Vector
 *               long dimx : dim of the State Vector
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY: fusion() and Dist_maha()    
 ***************/


void Kal_tempo(long dimf,long dimy,long dimx)

{

    if ((dimf > lastdimf) || (dimx > lastdimx) || (dimy > lastdimy))
    {
      free(temp);
      lastdimf = MAX(dimf, lastdimf);
      lastdimy = MAX(dimy, lastdimy);
      lastdimx = MAX(dimx, lastdimx);

      temp = (double *)malloc ( (lastdimf*(lastdimf+1)/2 + lastdimx*lastdimf +
	      lastdimf + lastdimf*lastdimy + lastdimf*lastdimx + lastdimf*(lastdimf+1)/2 +
	      lastdimx*lastdimf + lastdimx*lastdimx + lastdimx*lastdimf + 3*lastdimx +
	      lastdimx*lastdimx + lastdimx*(lastdimx+1)/2 )*sizeof(double));

/*       assert(temp); */
      gamma = temp;
      K = gamma + lastdimf*(lastdimf+1)/2;
      f = K + lastdimx*lastdimf;
      jacy = f + lastdimf;
      jacx = jacy + lastdimf*lastdimy;
      W = jacx + lastdimf*lastdimx;
      W1 = W + lastdimf*(lastdimf+1)/2;
      W2 = W1 + lastdimx*lastdimf;
      W3 = W2 + lastdimx*lastdimx;
      txpp = W3 + lastdimx*lastdimf;
      txll = txpp + lastdimx;
      xntp = txll + lastdimx;
      vxntp = xntp + lastdimx;
      vxntpp = vxntp + lastdimx*lastdimx;

    }
  }









