/*******************************************************
* This file deals with :
*     a) interfacing the nomadic matrix structure (called matrix form 
*     hereafter) defined in mat.h with the sgfull structure (sge form) 
*     of the package formed by the files "sgeco.c,sgefa.c,sgesl.c,blas.c
*     and ge.h"
*
*     b) using the package for inversing matrices.
*
*     Those functions are NOT intended to be called by the user of
*     the matrix library. The user should call the functions in "matrix_i.c"
*     which, in turn, will call this package if needed.
********************************************************/ 

#include <stdio.h>
#include <stdlib.h>

#include "mat.h"
#include "sge_prot.h"
#include "matsge_p.h"

static double *STAT_MX;
static int MX_SIZE = 0;

/******
Copy a matrix in matrix.c (matrix.h) into  a sge (ge.h)  matrix form
******/


/***************
 * FUNCTION:     mx2sge
 * PURPOSE:      Copy a matrix in matrix.c (matrix.h) into  a sge (ge.h) 
                 matrix form
 * ARGUMENTS:    double *m : input
 *               sgfull *sge : output
 *               long r : dimensions of m
 *               long c :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:      mx_trans (matrix.c)
 * CALLED BY:  mx_inv_sg()
 ***************/

void mx2sge(double *m,sgfull *sge,long r,long c)

{
 double *mt;
 long i;

 if(r*c > MX_SIZE){
   STAT_MX = Fallocmat(r,c);
   MX_SIZE = r*c;
 }
 mt = STAT_MX;

 /** we transpose since sgefull structure uses a collumn by collumn 
     representation **/

 mx_trans(m,mt,r,c);

 for(i = 0; i < c; i++)
   sge->pd[i]  = (mt + i*r);

 sge->cd = c;
 sge->rd = r;
 
}



/***************
 * FUNCTION:     sge2mx
 * PURPOSE:      copies a sge structure to a matrix structure
 * ARGUMENTS:    sgfull *sg : input
 *               double *m : output
 *               long r : dimension of m
 *               long c :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


void sge2mx(sgfull *sg,double *m,long *r,long *c)

{
 double *mt;

 *c = sg->cd;
 *r = sg->rd;
 mt = sg->pd[0];
 mx_trans(mt,m,*r,*c);
 
}



/***************
 * FUNCTION:     sgefree
 * PURPOSE:      freeing an sgfull structure. 
                 WARNING: this function assumes you used m2sge to fill up
                 the sgfull struture.
 * ARGUMENTS:    sgfull sg :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void sgefree(sgfull *sg)
    

{
  free(*(sg->pd));
  free(sg);
}

/*******************
inverts the matrix sg ; result in res in mx format

returns 0 if ok
returns 1 if bad conditionning
*******************/

/***************
 * FUNCTION:     sginvert
 * PURPOSE:      inverts an sgfull structure and puts the result in a matrix array.
 * ARGUMENTS:    sgfull sg : input
 *               double *res : output
 * ALGORITHM:    let us say sg is of dimension n*n
                 calls sgeco for putting "sg" under triangular form.
	         then, calls n times sgesl (solving sg.x = b) 
	         with b = [1,0...], then b = [0,1,0...] and so on .
                  
 * RETURN:       long : sgeco checks the matrix conditionning and returns
	                a condition number (called "test") in this function.
                        Here, we asks for test > 10e-13 in order to get
                        an accurate result.
	                sginvert returns 0 on success (good condition number)
                                 returns 1 on failure  
 * SIDE EFFECT:  
 * CALLS:        sgeco()/sgeco.c, sgesl()/sgesl.c
 * CALLED BY:    mx_inv_sg()
 ***************/

long sginvert(sgfull *sg,double *res)
     
{

  long r,c,i;
  long *iptv;
  double *x,*b,test;

  r = c = sg->rd;
  

  
  iptv = (long *)malloc( sizeof(long)*c);
  x = (double *)malloc(sizeof(double)*c);

  mx_init(res,c,c,0.);
 
  if(sgeco(sg,iptv,&test,x) != 0)
    return(1);

  if(test < 1e-13)
    return(1);

  for(i = 0; i < c; i ++){
    b = res + i*c ;
    b[i] = 1.;

    sgesl(sg,iptv,b,0);

  }

  free(iptv);
  free(x);
  return(0);

}


/***************
 * FUNCTION:     mx_inv_sg
 * PURPOSE:      inverts m1 into m2. 
                 can be done "on the spot".
 * ARGUMENTS:    double *m1 :input 
 *               double *m2 :output
 *               long r :dimensions of m1 and m2
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        mx2sge(), sginvert(), mx_trans()/matrix.c
 * CALLED BY:    mx_inv()/matrix_i.c, mxs_inv()/matrix_i.c
 ***************/


long mx_inv_sg(double *m1,double *m2,long r,long c)
     
{
  
  sgfull sg;

  mx2sge(m1,&sg,r,c);

  if( sginvert(&sg,m2) != 0)
    return(1);

  mx_trans(m2,m2,r,c);

  return(0);

}

  

  
  

  
  

  



