  /***************************************************************
 ******************************************************************
****	    Gaussian Elimination with partial pivoting		****
**** 		 and condition number estimation.		****
****	  This file contains the factorization driver and 	****
****	      condition number estimation routine SGECO		****
 ******************************************************************
  ****************************************************************/


#include <math.h>


#include "sge_prot.h"

#define copysign(x,y) (y>0 ? x : -x)


long sgeco( struct FULL *a,long *ipvt,double  *rcond,double *z )

/*
  PURPOSE
      SGECO factors a real matrix by gaussian elimination
      and estimates the condition of the matrix.
 
  REMARKS
      If  rcond  is not needed, SGEFA is slightly faster.
      to solve  A*x = b , follow SGECO by SGESL.
      To compute  inverse(A)*c , follow SGECO by SGESL.
      
 
  INPUT
 	a	A pointer to the FULL matrix structure.  
		See the definition in ge.h.
 
  OUTPUT
	a	A pointer to the FULL matrix structure containing 
		an upper triangular matrix and the multipliers
		which were used to obtain it.
		The factorization can be written  a = l*u  where
		l  is a product of permutation and unit lower
		triangular matrices and  u  is upper triangular.
	ipvt	An integer vector (of length a->cd) of pivot indices.
	rcond   A double estimate of the reciprocal condition of  A .
                for the system  A*x = b , relative perturbations
                in  A  and  b  of size  epsilon  may cause
                relative perturbations in  x  of size  epsilon/rcond .
                If  rcond  is so small that the logical expression
                            1.0 + rcond .eq. 1.0
                is true, then  a  may be singular to working
                precision.  In particular,  rcond  is zero  if
                exact singularity is detected or the estimate
                underflows.
         z      A double vector (of length a->cd) for a work vector 
	        whose contents are usually unimportant.  If  A  is 
		close to a singular matrix, then  z  is an approx-
		imate null vector in the sense that:
		         norm(a*z) = rcond*norm(a)*norm(z) .

    RETURNS
		= -1  Matrix is not square.
		=  0  Normal return value.
		=  k  if  u(k,k) .eq. 0.0 .  This is not an error
		      condition for this subroutine, but it does
		      indicate that sgesl or sgedi will divide by zero
		      if called.  Use  rcond  in sgeco for a reliable
		      indication of singularity.

    ROUTINES
	SGEFA(), blas sasum() and sdot(), copysign(), fabs();

    WARNINGS
        This routine uses the UN*X math library routines 
	copysign() and fabs().
*/
{
    void saxpy(), sscal();
    register int j;
    long k, n, info, sgefa();
    register double s;
    double sasum(), sdot();
    double ek, anorm, ynorm;

    n = a->cd;

    /* Compute 1-norm of A. */
    for( j=0, anorm=0.0; j<n; j++ ) {
	register double sum;

	sum = sasum( n, a->pd[j], 1 );
	anorm = (double)sum > anorm ? (double)sum : anorm;
    }

    /* Factor A. */
    info = sgefa( a, ipvt );
 
    /*
     * rcond = 1/(norm(a)*(estimate of norm(inverse(a)))) .
     * estimate = norm(z)/norm(y) where  a*z = y  and  trans(a)*y = e .
     * trans(a)  is the transpose of a .  The components of  e  are
     * chosen to cause maximum local growth in the elements of w  where
     * trans(u)*w = e .  The vectors are frequently rescaled to avoid
     * overflow.
     */

    /* solve trans(u)*w = e */
    ek = 1.0;
    for( j=0; j<n; j++ ) {
	z[j] = 0.0e0;
    }
    for( k=0; k<n; k++ ) {
	register double zk = z[k];
	double wk, wkm, sm;
	int kp1 = k+1;

	if( zk != 0.0 ) ek = (double)copysign( (double)ek, (double)-zk );
	if( fabs( (double)(ek-zk) ) > fabs( (double)pelem(a,k,k) ) ) {
            s = (double)fabs( (double)pelem(a,k,k) )/(double)fabs( (double)(ek-zk) );
            sscal( n, (double)s, z, 1 );
	    zk = z[k];
            ek *= s;
	}
	wk = ek - zk;
	wkm = -ek - zk;
	s = (double)fabs( (double)wk );
	sm = (double)fabs( (double)wkm );
	if( pelem(a,k,k) == 0.0 ) {
            wk = 1.0;
            wkm = 1.0;
	}
	else {
            wk /= pelem(a,k,k);
            wkm /= pelem(a,k,k);
	}
	if( kp1<n ) {
	    for( j=kp1; j<n; j++ ) {
		sm = sm + (double)fabs( (double)(z[j]+wkm*pelem(a,k,j)) );
		z[j] += ((double)wk)*pelem(a,k,j);
		s += (double)fabs( (double)z[j] );
	    }
            if( s < sm ) {
		register double t = wkm-wk;

		wk = wkm;
		for( j=kp1; j<n; j++ ) {
		    z[j] += t*pelem(a,k,j);
		}
	    }
	}
	z[k] = wk;
    }
    s = 1.0/(double)sasum( n, z, 1 );
    sscal( n, (double)s, z, 1 );

    /* Solve trans(L)*y = w. */
    for( k=n-1; k>=0; k-- ) {
	register int   l;
	register double t;

	if( k < n-1 ) z[k] += (double)sdot( n-k-1, a->pd[k]+k+1, 1, (z+k+1), 1 );
	if( fabs( (double)z[k] ) > 1.0 ) {
            s = 1.0/(double)fabs( (double)z[k] );
            sscal( n, (double)s, z, 1 );
	}
	l = ipvt[k];
	t = z[l];
	z[l] = z[k];
	z[k] = t;
    }
    s = 1.0/(double)sasum( n, z, 1 );
    sscal( n, (double)s, z, 1);

    ynorm = 1.0;

    /* solve L*v = y. */
    for( k=0; k<n; k++ ) {
	register int   l;
	register double t;

	l = ipvt[k];
	t = z[l];
	z[l] = z[k];
	z[k] = t;
	if( k < n-1 ) saxpy( n-k-1, (double)t, (a->pd[k]+k+1), 1, (z+k+1), 1 );
	if( fabs( (double)z[k] ) > 1.0)  {
            s = 1.0/(double)fabs( (double)z[k] );
            sscal( n, (double)s, z, 1 );
            ynorm *= s;
	}
    }
    s = 1.0/(double)sasum( n, z, 1 );
    sscal( n, (double)s, z, 1 );
    ynorm *= s;

    /* Solve  U*z = v. */
    for( k=n-1; k>=0; k-- ) {
	register double t;

	if( fabs( (double)z[k] ) > fabs( (double)pelem(a,k,k) ) ) {
            s = (double)fabs( (double)pelem(a,k,k) )/(double)fabs( (double)z[k] );
            sscal( n, (double)s, z, 1 );
            ynorm *= s;
	}
	if( pelem(a,k,k) == 0.0 ) z[k] = 1.0;
	else z[k] /= pelem(a,k,k);
	t = -z[k];
	saxpy( k, (double)t, a->pd[k], 1, z, 1 );
    }

    /* Make znorm = 1.0. */
    s = 1.0/(double)sasum( n, z, 1 );
    sscal( n, (double)s, z, 1 );
    ynorm *= s;

    if( anorm == 0.0e0)  *rcond = 0.0;
    else *rcond = ynorm/anorm;

    return( info );
}
