/********************************************************************
  matrix.c

  Functions to deal with matrices:
  We will work with 2 different kinds of matrices:

(1) "full" matrix = non-symmetrical
    coefficients are stocked row by row; 

    

    Example :

    [ 1 2 3 ]
    [ 4 5 6 ]
    [ 7 8 9 ]

    is represented as [1 2 3 4 5 6 7 8 9].
    We call it a f_mat

(2) Symetrical matrix : only ( n*(n+1)/ 2 coeffs are represented.
    They are stocked collumn by column

    Example :

    [ 1 2 4 ]
    [ 2 3 5 ]  
    [ 4 5 6 ]

    is represented as [1 2 3 4 5 6].
    We call it a s_mat

   NOTE: there is a function called mxs_fill allowing you to fill up
	 an s_mat in a more intuitive maner.
	 for example, for filling the matrix above, you would do:
              mxs_fill(s_mat,3,3,1,2,4,3,5,6)

  Access to the elements:
  **********************
     In addition to the functions in this file, the header file "mat.h"
     provides two macros for accessing the elements in an intuitive 
     way:

     

  Naming Convention:
  *****************
	 -the common prefix of all function is mx

	 -In case of symmetrical or mixed functions, this prefix is
	  immediately followed by the type of matrices 
	  the function is dealing with. types are "f" for full and "s"
	  for symmetrical. Types appear in the same order as matrices appear
          in the input argument list. 
           
	 -after types, there is an underscore followed by the description
          of the function itself

  Parameters passing:
  ******************
         -the first parameters are the input matrices
         -right after the input matrices, comes the output matrix
         -the last parameters describe the size of each matrix, in the order
	  of appearance, row first then collumn.
	 -there can be extra parameters for functions initializing matrices
	  which constant values. (see mx_fill)

  Return values
  *************
	  Functions test the consistency between dimensions.
	  If OK 
	     return 0
          else 
             return 1 
    
	  for example:

	  mx_mult(m1,m2,m,r1,c1,r2,c2) will check that c1 = r2 . If not
          it returns 1, otherwise it returns 0.
  
  Examples:
  *********
              mxfs_mult(m1,m2,m,r1,c1,r2,c2)

         multiplies an r1xc1 full matrix m1  by an r2xc2 symmetrical matrix m2.
         The result is in m which has to be big enough (e.g r1xc2 in this
         case). Memory allocation is left to the user.
         
	     

         Unless specified, the result matrix can be one of the input
         matrix (result "on the spot"). Then, the call
 
              mxfs_mult(m1,m2,m1,r1,c1,r2,c2); 

         is perfectly legal as long as m1 is of the right size.

	 NOTE: even when a matrix is supposed to be symmetrical, you still have
         to give both the row  and the collumn dimensions. This is
         a little bit stupid, though. Probably to be changed. 
        

  Available Functions:
  **********************

(1) on "full" matrices :
    ------------------

- mx_fill    (m, r, c, coeff1, coeff2, ...)    fills m with coeffs
- mx_init (mat, r, c, val)                     initialize all coeffs of  m at "val"

- mx_copy    (m1, m2, r, c)           --->  m2 = m1
- mx_abs     (m1, m2, r, c)           --->  m2 = absolute value (m1)
- mx_mulscal (m1, real, m2, r, c) --->  m2 = real . m1
- mx_trans   (m1, m2, r, c)           --->  m2 = transpose(m1)
- mx_plus    (m1, m2, m, r, c)        --->  m = m1 + m2
- mx_plustr  (m1, m, r, c)            --->  m = m1 + transpose(m1) (W: m is sym)
- mx_diff    (m1, m2, m, r, c)        --->  m = m1 - m2

- mx_mult (m1, m2, m, r1, c1, r2, c2) --->  m = m1 * m2
- mx_trmult1 (m1, m2, r, c)           --->  m2 = transpose(m1)* m1
- mx_trmult2 (m1, m2, r, c)           --->  m2 = m1 * transpose(m1)

Remark : mx_trmult1 and mx_trmult2 give symetrical results 

- mx_inf     (m1, m2, r, c)          --->  returns 1 if, 
                                             for all i, m1[i] < m2[i]
                                           returns 0 otherwise

- mx_diff_abs_inf (m1, m2, m, r, c)  --->  returns 1 if, 
                                             for all i, abs(m1[i]-m2[i]) < m[i]
                                           returns 0 otherwise

- mx_inv(m1,m2,r,c)                  ---> m2 = inverse(m1)

- mx_lsqr(m,x,b,r,c) :               ---> Solves the system m.x = b of r equations 
                                          with c unknown using least square;
(2) symetrical matrices :
    --------------------

- mxs_fill    (m, r, c, coeff1, coeff2, ...)    fills m with coeffs
- mxs_init (mat, r, c, val)                     initialize all coeffs of  m at "val"

- mxs_copy    (m1, m2, r, c)           --->  m2 = m1
- mxs_abs     (m1, m2, r, c)           --->  m2 = absolute value (m1)
- mxs_mulscal (m1, real, m2, r, c) --->  m2 = real . m1
- mxs_plus    (m1, m2, m, r, c)        --->  m = m1 + m2

- mxs_diff    (m1, m2, m, r, c)        --->  m = m1 - m2

- mxs_mult (m1, m2, m, r1, c1, r2, c2) --->  m = m1 * m2
                                             WARNING: mxs_mult produces a f_matrix !! 
                                             m has to be allocated accordingly

- mxs_inf     (m1, m2, r, c)          --->  returns 1 if, 
                                             for all i, m1[i] < m2[i]
                                           returns 0 otherwise

- mxs_diff_abs_inf (m1, m2, m, r, c)  --->  returns 1 if, 
                                             for all i, abs(m1[i]-m2[i]) < m[i]
                                           returns 0 otherwise

- mx_inv(m1,m2,r,c)                  ---> m2 = inverse(m1)

- mxs_fill    (m, r, c, coeff1, coeff2, ...)    
- mxs_init (mat, r, c, val)                     
- mxs_copy    (m1, m2, r, c)           --->   m2 = m1
- mxs_plus    (m1, m2, m, r, c)        --->   m = m1 + m2
- mxs_diff    (m1, m2, m, r, c)        --->   m = m1 - m2
- mxs_mulscal (m1, real, m2, r, c) --->   m2 = real . m1
- mxs_mult (m1, m2, m, r1, c1, r2, c2) --->   m = m1 * m2  (WARNING: m is FULL)
- mxs_inf     (m1, m2, r, c)          --->  (m1 < m2 ? 1 : 0)
- mxs_diff_abs_inf (m1, m2, m, r, c)  --->  (absolute value(m1-m2) < z ? 1 : 0)
- mxs_inv(m1,m2,r,c)                  ---> m2 = inverse(m1)
  

(3) Full and Symetrical:
    --------------------

- mxfs_copy    (m1, m2, r, c)           --->   m2(sym) = m1(full)
- mxsf_copy    (m1, m2, r, c)           --->   m2(full) = m1(sym)
- mxsf_plus    (m1, m2, m, r, c)        --->   m = m1 + m2
- mxfs_plus    (m1, m2, m, r, c)        --->   m = m1 + m2
- mxsf_diff    (m1, m2, m, r, c)        --->   m = m1 - m2
- mxfs_diff    (m1, m2, m, r, c)        --->   m = m1 - m2
- mxsf_mult (m1, m2, m, r1, c1, r2, c2) --->   m = m1 * m2
- mxfs_mult (m1, m2, m, r1, c1, r2, c2) --->   m = m1 * m2
- mxsf_multtr (m1,m2,m, r1, c1, r2, c2) --->   m = m1 * m2_tr
- mxfs_trmult (m1,m2,m, r1, c1, r2, c2) --->   m = m1_tr * m2

- mxfs_trmultm (m1, m2, m, r1, c1, r2, c2)  ---> m = transpose(m1) . m2. m1
- mxfs_mmulttr (m1, m2, m, r1, c1, r2, c2)  --->  m = m1 . m2. transpose(m1)

________________________________________________________________*/

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include "mat.h"


#define VERYSMALL(x) ( (x) > -1e-15 && (x) < 1e-15 )

#define Falloctemp(l,c) ( F_alloctemp ((l) * (c)) )
#define Salloctemp(l,c) ( F_alloctemp ((l) * ((c) + 1) / 2) )


/**********************************************************************

       Static Global variables for temporary computations
         
**********************************************************************/

static long dim_tempo = 0;
static double *tempo;

double *F_alloctemp(long n);





/***********************************************************************

  mx_fill (mat, r, c, coeff1, coeff2, ..., coeffn)
  
----------------------------------------------------------------------*/

/***************
 * FUNCTION:     mx_fill
 * PURPOSE:      fills up the matrix "mat" with a variable number
                 of values. 
 * ARGUMENTS:    double *mat,... :f_mat
                 variable list: first argument = matrix to be filled
		                second and third: row and collumn dimensions
				                  of mat
                                list of (row x collumn) doubles to fill
                                "mat" with
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long mx_fill (double *mat,...)


{

  va_list ap;
  long r, c, i;
  

  va_start (ap,mat);

  
  r = va_arg (ap, long);
  c = va_arg (ap, long);

  for (i = 0; i < Fmsize(r, c); i++)
      *mat++ = (double) va_arg (ap, double);

  va_end (ap);
  return (0);
}



/***************
 * FUNCTION:     mx_init
 * PURPOSE:      setting all values of "m" to be "val"
 * ARGUMENTS:    double *m : f_matrix to be initialized
 *               long r : row dimension
 *               long c : collumn
 *               double val : value 
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long mx_init (double *m,long r,long c,double val)

{
  long i;
  
  for (i = 0; i < Fmsize(r,c); i++)
    *m++ = val;

  return(0);
}









/***************
 * FUNCTION:     mx_copy
 * PURPOSE:      copy m1 onto m2 (full matrices)
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : f_matrix
 *               long r :
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mx_copy (double *m1,double *m2,long  r,long  c)

{
    mxcopy (m1, m2, Fmsize(r,c));
    return (0);
}


/***************
 * FUNCTION:     mx_abs
 * PURPOSE:      takes absolute values of each element of x
                 and copies them in ax .
 * ARGUMENTS:    double *x :
 *               double *ax :
 *               long r :
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mx_abs (double *x, double *ax, long r, long c)

{
  long k;
  for (k = 0; k < Fmsize(r, c); k++)
    *ax++ = fabs(*x++);
    	return(0);
}



/***************
 * FUNCTION:     mx_plus
 * PURPOSE:      m = m1 + m2 
 * ARGUMENTS:    double *m1 :
 *               double *m2 :
 *               double *m :
 *               long r :
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long mx_plus (double *m1, double *m2, double *m, long r, long c)


{
  long k;
  for (k = 0; k < Fmsize(r, c); k++)
      *m++ = *m1++ + *m2++;
  return (0);
}


/***************
 * FUNCTION:     mx_plustr
 * PURPOSE:      m = m1 + transpose(m1); 
                 WARNING: cannot be done on the spot.
                 m = s_matrix
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m : s_matrix
 *               long r :
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mx_plustr (double *m1,double *m,long  r,long  c)    

{
  long i,j;

  if (r != c)
    {
      printf("mx_plustr : non_square matrix !!!\n");
      return (1);
    }

  for (i = 0; i < r; i++)
    for (j = i; j < r; j++)
      *(m + j*(j+1)/2 + i) = Fmelt(m1, i,j,c) + Fmelt(m1, j,i,c);

  return (0);
}



/***************
 * FUNCTION:     mx_diff
 * PURPOSE:      m <-- m1 - m2
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : f_matrix
 *               double *m : f_matrix
 *               long r :
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/ 

long mx_diff (double *m1,double *m2, double *m, long r, long c)

{
  long k;
  for (k = 0; k < Fmsize(r, c); k++)
      *m++ = *m1++ - *m2++;
  return (0);
}






/***************
 * FUNCTION:     mx_mulscal
 * PURPOSE:      m2 <-- real.m1
 * ARGUMENTS:    double *m1 :f_matrix
 *               double real :
 *               double *m2 :f_matrix
 *               long r :
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mx_mulscal (double *m1,double real, double *m2,long  r, long c)

{
  long k;
  for (k = 0; k < Fmsize(r, c); k++)
      *m2++ = real * *m1++;
  return (0);
}


/***************
 * FUNCTION:     mx_trans
 * PURPOSE:      m2 <-- transpose(m1)
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : f_matrix
 *               long r :
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mx_trans (double *m1, double *m2,long  r,long c)

{
  double *temp,*aux;
  long i, j;


  if (m1 == m2)
    {
      /* allocating temporary memory */

      aux = temp = Falloctemp(r,c);
      for (i = 0; i < c; i++)
	for (j = 0; j < r; j++)
	  *temp++ = Fmelt (m1, j, i, c);

      /* copying over m2 */

      for(i = 0; i < c*r; i++)
	*m2++ = *aux++;
    }
  else
    {
      for (i = 0; i < c; i++)
	for (j = 0; j < r; j++)
	  *m2++ = Fmelt (m1, j, i, c);
    }
      
  return(0);
}


/***************
 * FUNCTION:     mx_mult
 * PURPOSE:      m <-- m1.m2
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : f_matrix
 *               double *m :  f_matrix
 *               long r1 :  dimensions of m1 
 *               long c1 :
 *               long r2 :  dimensions of m2
 *               long c2 :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mx_mult (double *m1,double *m2, double *m, long r1,long c1,
	      long r2,long c2)


{
  long i,j,k,flag = 0;
  double val;
  double *res, *aux;

  if (c1 != r2)
    {
      printf ("mx_mult : incompatible sizes\n");
      return (1);
    }

  if (m == m1 || m == m2)
    {
      res = Falloctemp (r1, c2);
      aux = res;
      flag = 1;
    }
  else
    res = m;

  for (i = 0; i < r1; i++)
    for (j = 0; j < c2; j++)
      {
	val = 0.;
	for (k = 0; k < c1; k++)
	  val += Fmelt (m1, i, k, c1) * Fmelt (m2, k, j, c2);
	*res++ = val;
      }

  if (flag)
    {
      res = aux;
      for (i = 0; i < Fmsize (r1, c2); i++)
	*m++ = *res++;
    }

  return (0);
}

/***************
 * FUNCTION:     mx_trmult
 * PURPOSE:      m <-- transpose(m1).m2
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : f_matrix
 *               double *m :  f_matrix
 *               long r1 :  dimensions of m1 
 *               long c1 :
 *               long r2 :  dimensions of m2
 *               long c2 :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/



long mx_trmult (double *m1,double  *m2,double *m,long r1,
		long  c1,long r2,long c2)


{
  long i,j,k,flag = 0;
  double val;
  double *res, *aux;

  if (r1 != r2)
    {
      printf ("mx_mult : incompatible sizes\n");
      return (1);
    }

  if (m == m1 || m == m2)
    {
      res = Falloctemp (c1, c2);
      aux = res;
      flag = 1;
    }
  else
    res = m;

  for (i = 0; i < c1; i++)
    for (j = 0; j < c2; j++)
      {
	val = 0.;
	for (k = 0; k < r1; k++)
	  val += Fmelt (m1, k, i, c1) * Fmelt (m2, k, j, c2);
	*res++ = val;
      }

  if (flag)
      mxcopy (aux, m, Fmsize (c1, c2));

  return (0);
}


/***************
 * FUNCTION:     mx_multtr
 * PURPOSE:      m <-- m1.transpose(m2)
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : f_matrix
 *               double *m :  f_matrix
 *               long r1 :  dimensions of m1 
 *               long c1 :
 *               long r2 :  dimensions of m2
 *               long c2 :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/




long mx_multtr (double *m1,double *m2,double  *m,
		long r1,long c1,long r2,long c2)


{
  long i,j,k,flag = 0;
  double val;
  double *res, *aux;

  if (c1 != c2)
    {
      printf ("mx_multtr : incompatible sizes\n");
      return (1);
    }

  if (m == m1 || m == m2)
    {
      res = Falloctemp (r1, r2);
      aux = res;
      flag = 1;
    }
  else
    res = m;

  for (i = 0; i < r1; i++)
    for (j = 0; j < r2; j++)
      {
	val = 0.;
	for (k = 0; k < c1; k++)
	  val += Fmelt (m1, i, k, c1) * Fmelt (m2, j, k, c2);
	*res++ = val;
      }

  if (flag)
      mxcopy (aux, m, Fmsize (r1, r2));

  return (0);
}


/***************
 * FUNCTION:     mx_trmult1
 * PURPOSE:      m2 <-- transpose(m1).m1
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : s_matrix
 *               long r :  dimensions of m1 
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long mx_trmult1 (double *m1,double  *m2,long r,long c)

{
  long i,j,k,flag = 0;
  double val;
  double *res, *aux;

  if (m2 == m1)
    {
      if (r != c)
	{
	  printf ("mx_trmult1 : non-square matrix \n");
	  return (1);
	}
      res = Salloctemp (c, c);
      aux = res;
      flag = 1;
    }
  else
    res = m2;

  for (j = 0; j < c; j++)
    for (i = 0; i <= j; i++)
      {
	val = 0.;
	for (k = 0; k < r; k++)
	  val += Fmelt (m1, k, i, c) * Fmelt (m1, k, j, c);
	*res++ = val;
      }

  if (flag)
    {
      res = aux;
      for (i = 0; i < Smsize (c, c); i++)
	*m2++ = *res++;
    }

  return (0);
}


/***************
 * FUNCTION:     mx_trmult2
 * PURPOSE:      m2 <-- m1.transpose(m1)
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : s_matrix
 *               long r1 :  dimensions of m1 
 *               long c1 :
 *               long r2 :  dimensions of m2
 *               long c2 :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/



long mx_trmult2 (double *m1, double *m2,long r,long c)
{
  long i,j,k,flag = 0;
  double val;
  double *res, *aux;

  if (m2 == m1)
    {
      if (r != c)
	{
	  printf ("mx_trmult1 : non-square matrix\n");
	  return (1);
	}

      res = Salloctemp (r, r);
      aux = res;
      flag = 1;
    }
  else
    res = m2;

  for (j = 0; j < r; j++)
    for (i = 0; i <= j; i++)
      {
	val = 0.;
	for (k = 0; k < c; k++)
	  val += Fmelt (m1, i, k, c) * Fmelt (m1, j, k, c);
	*res++ = val;
      }

  if (flag)
    {
      res = aux;
      for (i = 0; i < Smsize (r, r); i++)
	*m2++ = *res++;
    }

  return (0);
}


/***************
 * FUNCTION:     mx_inf
 * PURPOSE:      returns 1 if all elements of x are smaller than the same 
                 element of y.
		 returns 0 otherwise.
 * ARGUMENTS:    double *x : f_matrix
 *               double *y : f_matrix
 *               long r : dimensions of x and y
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long mx_inf (double *x, double *y, long r, long c)

{
  long k;
  for (k = 0; (k<Fmsize(r, c) && (*x++ < *y++)); k++);
  return ( k<Fmsize(r, c) ? 0 : 1);
}


/***************
 * FUNCTION:     mx_diff_abs_inf
 * PURPOSE:      returns 1 if, 
                     for all i, abs(m1[i]-m2[i]) < m[i]
                 returns 0 otherwise
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : f_matrix
 *               double *m : f_matrix
 *               long r : dimensions of m1,m2 and m
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long mx_diff_abs_inf (double *m1,double *m2,double *m,long  r,long c)

{
  long k;
  for (k=0; (k<Fmsize(r, c) && fabs(*m1++ - *m2++) < *m++); k++);
  return ( k<Fmsize(r, c) ? 0 : 1);
}



/***************
 * FUNCTION:     mxs_fill
  * PURPOSE:     fills up the matrix "mat" with a variable number
                 of values. For example, filling up "m" as following
 
                                     [ 1 2 4 ]
                                     [ 2 3 5 ]  
                                     [ 4 5 6 ]

		 is done by mxs_fill(m,3,3,1,2,4,3,5,6)
 
 * ARGUMENTS:    double *mat,... :s_mat
                 variable list: first argument = matrix to be filled
		                second and third: row and collumn dimensions
				                  of mat
                                list of (row x (row +1)/2) doubles to fill
                                "mat" with.
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/





long mxs_fill (double *mat,...)


{
  va_list ap;
  long r, c, i, j;
  

  va_start (ap,mat);

  
  r = va_arg (ap, long);
  c = va_arg (ap, long);

  if (r != c)
    {
      printf ("mxs : non-square symmetrical matrix !!!\n");
      return (1);
    }

  for (i = 0; i < r; i++)
    for (j = i; j < c; j++)
      *(mat + (j*(j+1))/2 + i) = (double) va_arg (ap, double);

  va_end (ap);
  return (0);
}


/***************
 * FUNCTION:     mxs_init
 * PURPOSE:      for all i, m[i] = val
 * ARGUMENTS:    double *m : s_matrix
 *               long r :  dimensions of m 
 *               long c :
 *               double val :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mxs_init (double *m,long r,long c,double val)

{
  long i;
  
  if (r != c)
    {
      printf ("mxs_init : non-square symmetrical matrix\n");
      return (1);
    }

  for (i = 0; i < Smsize(r,c); i++)
    *m++ = val;

  return(0);
}


/***************
 * FUNCTION:     mxs_copy
 * PURPOSE:      m2 <-- m1
 * ARGUMENTS:    double *m1 : s_matrix
 *               double *m2 : s_matrix
 *               long r :  dimensions of m1 and m2 
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/



long mxs_copy (double *m1,double *m2,long r,long c)


{
  

  if (r != c)
    {
      printf ("mxs_copy : non-square symmetrical matrix\n");
      return (1);
    }

  mxcopy (m1, m2, Smsize(r,c));
  return (0);
}


/***************
 * FUNCTION:     mxs_plus
 * PURPOSE:      m <-- m1 + m2
 * ARGUMENTS:    double *m1 : s_matrix
 *               double *m2 : s_matrix
 *               double *m :  s_matrix
 *               long r :  dimensions of m1, m2 and m 
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long mxs_plus (double *m1, double *m2, double *m, long r, long c)

{
  long k;

  if (r != c)
    {
      printf ("mxs_plus : non-square symetrical matrix !!!\n");
      return (1);
    }

  for (k = 0; k < Smsize(r, c); k++)
      *m++ = *m1++ + *m2++;
  return (0);
}


/***************
 * FUNCTION:     mxs_diff
 * PURPOSE:      m <-- m1 - m2
 * ARGUMENTS:    double *m1 : s_matrix
 *               double *m2 : s_matrix
 *               double *m :  s_matrix
 *               long r :  dimensions of m1, m2 and m 
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/




long mxs_diff (double *m1, double *m2, double *m, long r,long c)

{
  long k;

  if (r != c)
    {
      printf ("mxs_diff : non-square symetrical matrix !!!\n");
      return (1);
    }

  for (k = 0; k < Smsize(r, c); k++)
      *m++ = *m1++ - *m2++;
  return (0);
}


/***************
 * FUNCTION:     mxs_mulreal
 * PURPOSE:      m2 <-- real.m
 * ARGUMENTS:    double *m1 : s_matrix
 *               double real:
 *               double *m2 : s_matrix
 *               long r :  dimensions of m1 and m2 
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long mxs_mulreal (double *m1,double real, double *m2, long r,long c)

{
  long k;

  if (r != c)
    {
      printf ("mxs_mulreal : non-square matrix !!!\n");
      return (1);
    }

  for (k = 0; k < Smsize(r, c); k++)
      *m2++ = real * *m1++;
  return (0);
}

/***************
 * FUNCTION:     mxs_mult
 * PURPOSE:      m <-- m1.m2
 * ARGUMENTS:    double *m1 : s_matrix
 *               double *m2 : s_matrix
 *               double *m :  f_matrix
 *               long r1 :  dimensions of m1 
 *               long c1 :
 *               long r2 :  dimensions of m2
 *               long c2 :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 
    
***************/


long mxs_mult (double *m1, double *m2, double *m,
	       long r1,long c1,long r2,long c2)

{
  long i,j,k, flag = 0;
  double val;
  double *res, *aux;

  if (c1 != r1 || c2 != r2 || c1 != c2)
    {
      printf ("mxs_mult : incompatible sizes\n");
      return (1);
    }

  if (m == m1 || m == m2)
    {
      res = Falloctemp (r1, c2);
      aux = res;
      flag = 1;
    }
  else
    res = m;

  for (i = 0; i < r2; i++)
    for (j = 0; j < c2; j++)
      {
	val = 0.;
	for (k = 0; k < c1; k++)
	  val += Smelt (m1, i, k) * Smelt (m2, k, j);
	*res++ = val;
      }

  if (flag)
      mxcopy (aux, m, Fmsize (r1, c2));

  return (0);
}



/***************
 * FUNCTION:     mxs_inf
 * PURPOSE:      returns 1 if each element of x is smaller than the same 
                 element of y.
		 returns 0 otherwise.
 * ARGUMENTS:    double *x : s_matrix
 *               double *y : s_matrix
 *               long r : dimensions of x and y
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:
 ***************/



long mxs_inf (double *x, double *y, long r, long c)

{
  long k;
  for (k = 0; (k<Smsize(r, c) && (*x++ < *y++)); k++);
  return ( k<Smsize(r, c) ? 0 : 1);
}



/***************
 * FUNCTION:     mxs_diff_abs_inf
 * PURPOSE:      returns 1 if, 
                     for all i, abs(m1[i]-m2[i]) < m[i]
                 returns 0 otherwise
 * ARGUMENTS:    double *m1 : s_matrix
 *               double *m2 : s_matrix
 *               double *m : s_matrix
 *               long r : dimensions of m1,m2 and m
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mxs_diff_abs_inf (double *m1, double *m2, double *m,long r,long c)

{
  long k;
  for (k=0; (k< r * (r + 1) / 2 && fabs(*m1++ - *m2++) < *m++); k++);
  return ( k<Fmsize(r, c) ? 0 : 1);
}


/***************
 * FUNCTION:     mxfs_copy
 * PURPOSE:      m2 <-- m1
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : s_matrix
 *               long r :  dimensions of m1 and m2 
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mxfs_copy (double *m1, double *m2, long r, long c)

{
  long i,j;

  if (r != c)
    {
      printf ("mxs_copy : non-square matrix\n");
      return (1);
    }

  for (j = 0; j < c; j++)
    for (i = 0; i <= j; i++)
      *m2++ = Fmelt (m1,i,j,c);

  return (0);
}

/***************
 * FUNCTION:     mxsf_copy
 * PURPOSE:      m2 <-- m1
 * ARGUMENTS:    double *m1 : s_matrix
 *               double *m2 : f_matrix
 *               long r :  dimensions of m1 and m2 
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long mxsf_copy (double *m1, double *m2,long r,long c)

{
  long i,j;

  if (r != c)
    {
      printf ("mxsf_copy : non-square matrix\n");
      return (1);
    }

  for (j = 0; j < c; j++)
    for (i = 0; i < r; i++)
      *m2++ = Smelt (m1,i,j);

  return(0);
}




/***************
 * FUNCTION:     mxfs_plus
 * PURPOSE:      m <-- m1 + m2
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : s_matrix
 *               double *m :  f_matrix
 *               long r :  dimensions of m1, m2 and m 
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/
  
long mxfs_plus (double *m1, double *m2,double *m, long r, long c)

{
  long i,j;

  if (r != c)
    {
      printf ("mxfs_plus : non-square matrix !!!\n");
      return (1);
    }

  for (i = 0; i < r; i++)
    for (j = 0; j < c; j++)
      *m++ = *m1++ + Smelt (m2, i, j);
  return (0);
}

/***************
 * FUNCTION:     mxsf_plus
 * PURPOSE:      m <-- m1 + m2
 * ARGUMENTS:    double *m1 : s_matrix
 *               double *m2 : f_matrix
 *               double *m :  f_matrix
 *               long r :  dimensions of m1, m2 and m 
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

#define mxsf_plus (m1, m2, m, l, c) (mxfs_plus (m2, m1, m, l, c))



/***************
 * FUNCTION:     mxfs_diff
 * PURPOSE:      m <-- m1 - m2
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : s_matrix
 *               double *m :  f_matrix
 *               long r :  dimensions of m1, m2 and m 
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mxfs_diff (double *m1, double *m2, double *m,long r,long c)

{
  long i,j;

  if (r != c)
    {
      printf ("mxs : non-square matrix !!!\n");
      return (1);
    }

  for (i = 0; i < r; i++)
    for (j = 0; j < c; j++)
      *m++ = *m1++ - Smelt (m2, i, j);
  return (0);
}

/***************
 * FUNCTION:     mxsf_diff
 * PURPOSE:      m <-- m1 - m2
 * ARGUMENTS:    double *m1 : s_matrix
 *               double *m2 : f_matrix
 *               double *m :  f_matrix
 *               long r :  dimensions of m1, m2 and m 
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/



long mxsf_diff (double *m1, double *m2, double *m,long r,long c)

{
  long i,j;

  if (r != c)
    {
      printf ("mxs : non-square matrix !!!\n");
      return (1);
    }

  for (i = 0; i < r; i++)
    for (j = 0; j < c; j++)
      *m++ = Smelt (m1, i, j) - *m2++;
  return (0);
}


/***************
 * FUNCTION:     mxsf_mult
 * PURPOSE:      m <-- m1.m2
 * ARGUMENTS:    double *m1 : s_matrix
 *               double *m2 : f_matrix
 *               double *m :  f_matrix
 *               long r1 :  dimensions of m1 
 *               long c1 :
 *               long r2 :  dimensions of m2
 *               long c2 :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mxsf_mult (double *m1,double *m2, double *m,
		long r1,long c1,long r2,long c2)


{
  long i,j,k, flag = 0;
  double val;
  double *res, *aux;

  if (c1 != r1 || c1 != r2 || m == m1)
    {
      printf ("mxsf_mult : incompatible sizes\n");
      return (1);
    }

  if (m == m2)
    {
      res = Falloctemp (r1, c2);
      aux = res;
      flag = 1;
    }
  else
    res = m;

  for (i = 0; i < r1; i++)
    for (j = 0; j < c2; j++)
      {
	val = 0.;
	for (k = 0; k < c1; k++)
	  val += Smelt (m1, i, k) * Fmelt (m2, k, j, c2);
	*res++ = val;
      }

  if (flag)
    {
      res = aux;
      for (i = 0; i < Fmsize (r1, c2); i++)
	*m++ = *res++;
    }

  return (0);
}


/***************
 * FUNCTION:     mxsf_multtr
 * PURPOSE:      m <-- m1.transpose(m2)
 * ARGUMENTS:    double *m1 : s_matrix
 *               double *m2 : f_matrix
 *               double *m :  f_matrix
 *               long r1 :  dimensions of m1 
 *               long c1 :
 *               long r2 :  dimensions of m2
 *               long c2 :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/



long mxsf_multtr (double *m1, double *m2, double *m, 
		  long r1,long  c1,long r2,long c2)

{
  long i,j,k, flag = 0;
  double val;
  double *res, *aux;

  if (c1 != r1 || c1 != c2 || m == m1)
    {
      printf ("mxsf_multtr : incompatible sizes\n");
      return (1);
    }

  if (m == m2)
    {
      res = Falloctemp (r1, r2);
      aux = res;
      flag = 1;
    }
  else
    res = m;

  for (i = 0; i < r1; i++)
    for (j = 0; j < r2; j++)
      {
	val = 0.;
	for (k = 0; k < c1; k++)
	  val += Smelt (m1, i, k) * Fmelt (m2, j, k, c2);
	*res++ = val;
      }

  if (flag)
      mxcopy (aux, m, Fmsize (r1, r2));

  return (0);
}



/***************
 * FUNCTION:     mxfs_mult
 * PURPOSE:      m <-- m1.m2
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : s_matrix
 *               double *m :  f_matrix
 *               long r1 :  dimensions of m1 
 *               long c1 :
 *               long r2 :  dimensions of m2
 *               long c2 :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mxfs_mult (double *m1,double *m2, double *m, 
		long r1,long c1,long r2,long c2)


{
  long i,j,k, flag = 0;
  double val;
  double *res, *aux;

  if (c2 != r2 || c1 != r2 || m == m2)
    {
      printf ("mxfs_mult : incompatible sizes\n");
      return (1);
    }

  if (m == m1)
    {
      res = Falloctemp (r1, c2);
      aux = res;
      flag = 1;
    }
  else
    res = m;

  for (i = 0; i < r1; i++)
    for (j = 0; j < c2; j++)
      {
	val = 0.;
	for (k = 0; k < c1; k++)
	  val += Fmelt (m1, i, k, c1) * Smelt (m2, k, j);
	*res++ = val;
      }

  if (flag)
    {
      res = aux;
      for (i = 0; i < Fmsize (r1, c2); i++)
	*m++ = *res++;
    }

  return (0);
}


/***************
 * FUNCTION:     mxfs_trmult
 * PURPOSE:      m <-- transpose(m1).m2
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : s_matrix
 *               double *m :  f_matrix
 *               long r1 :  dimensions of m1 
 *               long c1 :
 *               long r2 :  dimensions of m2
 *               long c2 :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long mxfs_trmult (double *m1, double *m2,double *m,
		  long r1,long c1,long r2,long c2)

{
  long i,j,k, flag = 0;
  double val;
  double *res, *aux;

  if (c2 != r2 || r1 != r2 || m == m2)
    {
      printf ("mxfs_trmult : incompatible sizes\n");
      return (1);
    }

  if (m == m1)
    {
      res = Falloctemp (c1, c2);
      aux = res;
      flag = 1;
    }
  else
    res = m;

  for (i = 0; i < c1; i++)
    for (j = 0; j < c2; j++)
      {
	val = 0.;
	for (k = 0; k < r1; k++)
	  val += Fmelt (m1, k, i, c1) * Smelt (m2, k, j);
	*res++ = val;
      }

  if (flag)
      mxcopy (aux, m, Fmsize (c1,c2));

  return (0);
}



/***************
 * FUNCTION:     mxcopy
 * PURPOSE:      copies the array src over dest. 
 * ARGUMENTS:    double *src : array source
 *               double *dest : array destination
 *               long size : size of src and dest
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


void mxcopy(double *src, double *dest,long size)
{
long i;
for (i = 0; i < size; i++)
	*dest++ = *src++;
}


/***************
 * FUNCTION:     mxfs_trmultm
 * PURPOSE:      m <-- transpose(m1).m2.m1 (m2 is symmetrical
                 then the result is an s_matrix)
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : s_matrix
 *               double *m :  s_matrix
 *               long r1 :  dimensions of m1 
 *               long c1 :
 *               long r2 :  dimensions of m2
 *               long c2 :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mxfs_trmultm (double *m1,double *m2, double *m,
		   long r1,long c1,long r2,long c2)



{
  double *temp, *res, *aux, *aux2;
  long i,j,k, flag = 0;
  double val;

  if (r2 != c2 || r1 != r2)
    {
      printf ("mxsf_trmultm : incompatible sizes \n");
      return (1);
    }

  if (m == m1 || m == m2)
    {
      temp = F_alloctemp(c1*c2+ c1*(c1+1)/2);
      aux = temp;
      res = temp + c1*c2;
      aux2 = res;
      flag = 1;
    }
  else
    {
      temp = Falloctemp (c1, c2);
      aux = temp; 
      res = m;
    }

  for (i = 0; i < c1; i++)
    for (j = 0; j < c2; j++)
      {
	val = 0.;
	for (k = 0; k < r1; k++)
	  val += Fmelt (m1, k, i, c1) * Smelt (m2, k, j);
	*temp++ = val;
      }
  temp = aux;
  
  for (j = 0; j < c1; j++)
    for (i = 0; i <= j; i++)
      {
	val = 0.;
	for (k = 0; k < c2; k++)
	  val += Fmelt (temp, i, k, c2) * Fmelt (m1, k, j, c1);
	*res++ = val;
      }

  if (flag)
    {
      res = aux2;
      for (i = 0; i < Smsize (c1, c1); i++)
	*m++ = *res++;
    }

  return (0);
}

/***************
 * FUNCTION:     mxfs_mmulttr
 * PURPOSE:      m <-- m1.m2.transpose(m1)
                 since m2 is symetrical, the result is an s_matrix
 * ARGUMENTS:    double *m1 : f_matrix
 *               double *m2 : s_matrix
 *               double *m :  s_matrix
 *               long r1 :  dimensions of m1 
 *               long c1 :
 *               long r2 :  dimensions of m2
 *               long c2 :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mxfs_mmulttr (double *m1, double *m2, double *m, 
		   long r1,long c1,long r2,long c2)

{

  double *temp, *res, *aux, *aux2;
  long i,j,k,flag = 0;
  double val;

  if (r2 != c2 || c1 != c2)
    {
      printf ("mxsf_mmulttr : incompatible sizes \n");
      return (1);
    }


  if (m == m1 || m == m2)
    {
      temp = F_alloctemp(r1*c2 + r1*(r1+1)/2);
      aux = temp;
      res = temp + r1*c2;
      aux2 = res;
      flag = 1;
    }
  else
    {
      temp = Falloctemp (r1, c2);
      aux = temp;
      res = m;
    }

  
  for (i = 0; i < r1; i++)
    for (j = 0; j < c2; j++)
      {
	val = 0.;
	for (k = 0; k < c1; k++)
	  val += Fmelt (m1, i, k, c1) * Smelt (m2, k, j);
	*temp++ = val;
      }
  temp = aux;
  
  for (j = 0; j < r1; j++)
    for (i = 0; i <= j; i++)
      {
	val = 0.;
	for (k = 0; k < c2; k++)
	  val += Fmelt (temp, i, k, c2) * Fmelt (m1, j, k, c1);
	*res++ = val;
      }

  if (flag)
    {
      res = aux2;
      for (i = 0; i < Smsize (r1, r1); i++)
	*m++ = *res++;
    }

  return (0);
}



/*****************************************************
  Standard output on the terminal
  mx_print, mxs_print   

*****************************************************/


/***************
 * FUNCTION:     mx_print
 * PURPOSE:      print a f_matrix on the standard output
 * ARGUMENTS:    double *m :
 *               long r :
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long mx_print (double *m,long r,long c)

{
  long i,j;
  for (i=0;i<r;i++)
    {
      printf (" [ ");
      for (j=0;j<c;j++)
	printf ("%.10e ",Fmelt (m,i,j,c));
      printf ("]\n");
   }
return(0);
}




/***************
 * FUNCTION:     mxs_print
 * PURPOSE:      prints a s_matrix on the standard output
 * ARGUMENTS:    double *m :
 *               long r :
 *               long c :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mxs_print (double *m,long r,long c)

{
  long i,j;

  for(i=0;i<r;i++)
    {
      printf (" [ ");
      for (j=0;j<r;j++)
	printf("%.10e ",Smelt (m,i,j));
      printf ("]\n");
    }
return(0);
}


/***************
 * FUNCTION:     F_alloctemp
 * PURPOSE:      allocates temporary memory when doing "on the spot"
                 computations.
                 NOTE: tempo is a variable local to this file. It is never 
		 freed but rather reused and expanded when needed.
		 WARNING: do not free the returned pointer !!!!!
 * ARGUMENTS:    long n :
 * ALGORITHM:    
 * RETURN:       double *
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


double *F_alloctemp (long n)

{
  if (n <= dim_tempo)
    return(tempo);
  else
    {
      if (dim_tempo != 0)
	free(tempo);
      else n += 20;
      dim_tempo = n;
      return (tempo = (double *)malloc(dim_tempo*sizeof(double)));
    }
}
      
	
      











